""" Utility functions

Defines a set of utility functions used throughout the code
to make it more readable.

Functions
---------
hasfunction(obj: any, function_name: str) -> bool
    Return True if the object has a field named function_name
    that is callable. Otherwise, return False.
isiterable(obj: any)
    Return True if the object is iterable. Else, return False.
as_list(obj: any)
    If the input is iterable, convert it to list. 
    Otherwise, wrap the input in a list.
get_kwarg_names(function: Callable)
    Return the names of the keyword arguments the function accepts.
"""

import inspect

from typing import Callable, List


def hasmethod(obj: any, method_name: str) -> bool:
    """Check if an object has a callable method named method_name.

    Parameters
    ----------
    obj
        The object to be checked.
    method_name
        The name of the method to look for.

    Returns
    -------
    bool
        True if the object has an attribute method_name, and that
        attribute is callable.

    """

    return hasattr(obj, method_name) and callable(getattr(obj, method_name, None))


def isiterable(obj: any) -> bool:
    """Check if the input is iterable.
    Note that this function does not capture all possible cases
    and is subject to change in the future if issues arise.

    Parameters
    ----------
    obj
        The object to check.

    Returns
    -------
    bool
        True if the object has __next__ defined.

    """

    return hasattr(obj, "__next__")


def as_list(obj: any) -> list:
    """Ensure the input is a list.
    If the input is iterable, convert it to a list,
    otherwise wrap the input in a list.

    Parameters
    ----------
    obj
        The object that will be made a list.

    Returns
    -------
    list
        The input as a list.

    """

    try:
        return list(obj)
    except TypeError:
        return [obj]


def get_kwarg_names(function: Callable) -> List[str]:
    """Retrieve the names of the keyword arguments.
    Retrieve the names of the keyword arguments accepted by `function`
    as a list of strings.

    Parameters
    ----------
    function
        The function to retrieve keyword argument names from.

    Returns
    -------
    List[str]
        The accepted keyword arguments as a list of strings.

    """

    try:
        argspec = inspect.getfullargspec(function)
    except TypeError:
        return []

    if argspec.varargs:
        return argspec.kwonlyargs or []
    else:
        return argspec.args or []


def kwarg_has_default(function: Callable, argument: str) -> bool:
    """Returns true if an argument has a default value.

    Parameters
    ----------
    function : Callable
        The function to check.
    argument : str
        Name of the argument

    Returns
    -------
    bool

    """
    args = get_kwarg_names(function)

    if argument not in args:
        return False

    defaults = inspect.getfullargspec(function).defaults or ()

    return len(args) - args.index(argument) <= len(defaults)


def safe_call(function, **kwargs):
    """Calls a function, using keyword arguments from a dictionary of arguments.

    If the function does not accept one of the argument provided, it will not be passed. Does not support
    non-keyword arguments.

    Parameters
    ----------
    function : Callable
        The function to call
    kwargs
        Key-value pairs to draw input arguments from.
    """

    keys = get_kwarg_names(function)

    input_arguments = {}
    for key in keys:
        if key in kwargs:
            input_arguments[key] = kwargs[key]

    return function(**input_arguments)



""" Standard models for neural networks.

Classes
-------
ModelFeature 
    Base model feature class.
Convolutional, convolutional
    Creates and compiles a convolutional neural network.
UNet, unet
    Creates and compiles a U-Net neural network.
RNN, rnn
    Creates and compiles a recurrent neural network.
"""

from deeptrack.losses import nd_mean_absolute_error
from deeptrack.features import Feature
from deeptrack.layers import as_block
from tensorflow.keras import models, layers, optimizers
import numpy as np
import warnings


def _compile(
    model: models.Model, *, loss="mae", optimizer="adam", metrics=[], **kwargs
):
    """Compiles a model.

    Parameters
    ----------
    model : keras.models.Model
        The keras model to interface.
    loss : str or keras loss
        The loss function of the model.
    optimizer : str or keras optimizer
        The optimizer of the model.
    metrics : list, optional
    """

    model.compile(loss=loss, optimizer=optimizer, metrics=metrics)
    return model


class Model(Feature):
    def __init__(self, model, **kwargs):
        self.model = model
        super().__init__(**kwargs)

    def __getattr__(self, key):
        # Allows access to the model methods and properties
        try:
            return getattr(super(), key)
        except AttributeError:
            return getattr(self.model, key)


class KerasModel(Model):
    def __init__(
        self,
        model,
        loss="mae",
        optimizer="adam",
        metrics=[],
        compile=True,
        add_batch_dimension_on_resolve=True,
        **kwargs
    ):

        if compile:
            model.compile(loss=loss, optimizer=optimizer, metrics=metrics)

        super().__init__(
            model,
            add_batch_dimension_on_resolve=add_batch_dimension_on_resolve,
            metrics=metrics,
            **kwargs
        )

    def get(self, image, add_batch_dimension_on_resolve, **kwargs):
        if add_batch_dimension_on_resolve:
            image = np.expand_dims(image, axis=0)

        return self.model.predict(image)

    def __call__(self, *args, **kwargs):
        return self.model(*args, **kwargs)


def LoadModel(path, compile_from_file=False, custom_objects={}, **kwargs):
    """Loads a keras model from disk.

    Parameters
    ----------
    path : str
        Path to the keras model to load.
    compile_from_file : bool
        Whether to compile the model using the loss and optimizer in the saved model. If false,
        it will be compiled from the arguments in kwargs (loss, optimizer and metrics).
    custom_objects : dict
        Dict of objects to use when loading the model. Needed to load a model with a custom loss,
        optimizer or metric.
    """
    model = models.load_model(
        path, compile=compile_from_file, custom_objects=custom_objects
    )
    return KerasModel(model, compile=not compile_from_file, **kwargs)


def FullyConnected(
    input_shape,
    dense_layers_dimensions=(32, 32),
    dropout=(),
    flatten_input=True,
    number_of_outputs=3,
    output_activation=None,
    dense_block="dense",
    **kwargs
):
    """Creates and compiles a fully connected neural network.

    A convolutional network with a dense top.

    Parameters
    ----------
    input_shape : tuple of ints
        Size of the images to be analyzed.
    dense_layers_dimensions : tuple of ints
        Number of units in each dense layer.
    flatten_input : bool
        Whether to add a flattening layer to the input
    number_of_outputs : int
        Number of units in the output layer.
    output_activation : str or keras activation
        The activation function of the output.
    dense_block
    loss : str or keras loss function
        The loss function of the network.

    Returns
    -------
    keras.models.Model
        Deep learning network
    """

    dense_block = as_block(dense_block)

    ### INITIALIZE DEEP LEARNING NETWORK
    input_layer = layers.Input(shape=input_shape)

    layer = input_layer
    if flatten_input:
        layer = layers.Flatten()(layer)

    # DENSE TOP
    for dense_layer_number, dense_layer_dimension in zip(
        range(len(dense_layers_dimensions)), dense_layers_dimensions
    ):

        if dense_layer_number is 0 and not flatten_input:
            layer = dense_block(dense_layer_dimension, input_shape=input_shape)(layer)
        else:
            layer = dense_block(dense_layer_dimension)(layer)

        if dropout:
            layer = layers.Dropout(dropout[0])(layer)
            dropout = dropout[1:]

    # OUTPUT LAYER

    output_layer = layers.Dense(number_of_outputs, activation=output_activation)(layer)

    model = models.Model(input_layer, output_layer)

    return KerasModel(model, **kwargs)


def Convolutional(
    input_shape=(51, 51, 1),
    conv_layers_dimensions=(16, 32, 64, 128),
    dense_layers_dimensions=(32, 32),
    steps_per_pooling=1,
    dropout=(),
    dense_top=True,
    number_of_outputs=3,
    output_activation=None,
    output_kernel_size=3,
    loss=nd_mean_absolute_error,
    convolution_block="convolutional",
    pooling_block="pooling",
    dense_block="dense",
    **kwargs
):
    """Creates and compiles a convolutional neural network.
    A convolutional network with a dense top.
    Parameters
    ----------
    input_shape : tuple of ints
        Size of the images to be analyzed.
    conv_layers_dimensions : tuple of ints
        Number of convolutions in each convolutional layer.
    dense_layers_dimensions : tuple of ints
        Number of units in each dense layer.
    dropout : tuple of float
        Adds a dropout between the convolutional layers
    number_of_outputs : int
        Number of units in the output layer.
    output_activation : str or keras activation
        The activation function of the output.
    loss : str or keras loss function
        The loss function of the network.
    layer_function : Callable[int] -> keras layer
        Function that returns a convolutional layer with convolutions
        determined by the input argument. Can be use to futher customize the network.
    Returns
    -------
    keras.models.Model
        Deep learning network
    """

    # Update layer functions
    dense_block = as_block(dense_block)
    convolution_block = as_block(convolution_block)
    pooling_block = as_block(pooling_block)

    ### INITIALIZE DEEP LEARNING NETWORK

    if isinstance(input_shape, list):
        inputs = [layers.Input(shape for shape in input_shape)]
        inputs = layers.Concatenate(axis=-1)(inputs)
    else:
        network_input = layers.Input(input_shape)
        inputs = network_input

    layer = inputs

    ### CONVOLUTIONAL BASIS
    for conv_layer_dimension in conv_layers_dimensions:

        for _ in range(steps_per_pooling):
            layer = convolution_block(conv_layer_dimension)(layer)

        if dropout:
            layer = layers.SpatialDropout2D(dropout[0])(layer)
            dropout = dropout[1:]

        # add pooling layer
        layer = pooling_block(conv_layer_dimension)(layer)

    # DENSE TOP

    if dense_top:
        layer = layers.Flatten()(layer)
        for dense_layer_dimension in dense_layers_dimensions:
            layer = dense_block(dense_layer_dimension)(layer)
        output_layer = layers.Dense(number_of_outputs, activation=output_activation)(
            layer
        )
    else:

        output_layer = layers.Conv2D(
            number_of_outputs,
            kernel_size=output_kernel_size,
            activation=output_activation,
            padding="same",
            name="output",
        )(layer)

    model = models.Model(inputs, output_layer)

    return KerasModel(model, loss=loss, **kwargs)


# Alias for backwards compatability
convolutional = Convolutional


def UNet(
    input_shape=(None, None, 1),
    conv_layers_dimensions=(16, 32, 64, 128),
    base_conv_layers_dimensions=(128, 128),
    output_conv_layers_dimensions=(16, 16),
    dropout=(),
    steps_per_pooling=1,
    number_of_outputs=1,
    output_kernel_size=3,
    output_activation=None,
    loss=nd_mean_absolute_error,
    encoder_convolution_block="convolutional",
    base_convolution_block="convolutional",
    decoder_convolution_block="convolutional",
    output_convolution_block="convolutional",
    pooling_block="pooling",
    upsampling_block="deconvolutional",
    **kwargs
):

    """Creates and compiles a U-Net.
    Parameters
    ----------
    input_shape : tuple of ints
        Size of the images to be analyzed.
    conv_layers_dimensions : tuple of ints
        Number of convolutions in each convolutional layer during down-
        and upsampling.
    base_conv_layers_dimensions : tuple of ints
        Number of convolutions in each convolutional layer at the base
        of the unet, where the image is the most downsampled.
    output_conv_layers_dimensions : tuple of ints
        Number of convolutions in each convolutional layer after the
        upsampling.
    steps_per_pooling : int
        Number of convolutional layers between each pooling and upsampling
        step.
    number_of_outputs : int
        Number of convolutions in output layer.
    output_activation : str or keras activation
        The activation function of the output.
    loss : str or keras loss function
        The loss function of the network.
    layer_function : Callable[int] -> keras layer
        Function that returns a convolutional layer with convolutions
        determined by the input argument. Can be use to futher customize the network.
    Returns
    -------
    keras.models.Model
        Deep learning network.
    """

    # Update layer functions

    encoder_convolution_block = as_block(encoder_convolution_block)
    base_convolution_block = as_block(base_convolution_block)
    output_convolution_block = as_block(output_convolution_block)
    decoder_convolution_block = as_block(decoder_convolution_block)
    pooling_block = as_block(pooling_block)
    upsampling_block = as_block(upsampling_block)

    unet_input = layers.Input(input_shape)

    concat_layers = []

    layer = unet_input

    # Downsampling path
    for conv_layer_dimension in conv_layers_dimensions:
        for _ in range(steps_per_pooling):
            layer = encoder_convolution_block(conv_layer_dimension)(layer)
        concat_layers.append(layer)

        if dropout:
            layer = layers.SpatialDropout2D(dropout[0])(layer)
            dropout = dropout[1:]

        layer = pooling_block(conv_layer_dimension)(layer)

    # Bottleneck path
    for conv_layer_dimension in base_conv_layers_dimensions:
        layer = base_convolution_block(conv_layer_dimension)(layer)

    # Upsampling path
    for conv_layer_dimension, concat_layer in zip(
        reversed(conv_layers_dimensions), reversed(concat_layers)
    ):

        layer = upsampling_block(conv_layer_dimension)(layer)
        layer = layers.Concatenate(axis=-1)([layer, concat_layer])

        for _ in range(steps_per_pooling):
            layer = decoder_convolution_block(conv_layer_dimension)(layer)

    # Output step
    for conv_layer_dimension in output_conv_layers_dimensions:
        layer = output_convolution_block(conv_layer_dimension)(layer)

    output_layer = layers.Conv2D(
        number_of_outputs,
        kernel_size=output_kernel_size,
        activation=output_activation,
        padding="same",
    )(layer)

    model = models.Model(unet_input, output_layer)

    return KerasModel(model, loss=loss, **kwargs)


# Alias for backwards compatability
unet = UNet


def RNN(
    input_shape=(51, 51, 1),
    conv_layers_dimensions=(16, 32, 64, 128),
    dense_layers_dimensions=(32,),
    rnn_layers_dimensions=(32,),
    return_sequences=False,
    output_activation=None,
    number_of_outputs=3,
    **kwargs
):
    """Creates and compiles a recurrent neural network.

    Parameters
    ----------
    input_shape : tuple of ints
        Size of the images to be analyzed.
    conv_layers_dimensions : tuple of ints
        Number of convolutions in each convolutional layer during down-
        and upsampling.
    dense_layers_dimensions : tuple of ints
        Number of units in each dense layer.
    rnn_layers_dimensions : tuple of ints
        Number of units in each recurrent layer.
    number_of_outputs : int
        Number of convolutions in output layer.
    output_activation : str or keras activation
        The activation function of the output.
    loss : str or keras loss function
        The loss function of the network.

    Returns
    -------
    keras.models.Model
        Deep learning network.
    """

    ### INITIALIZE DEEP LEARNING NETWORK
    network = models.Sequential()

    ### CONVOLUTIONAL BASIS
    for conv_layer_number, conv_layer_dimension in zip(
        range(len(conv_layers_dimensions)), conv_layers_dimensions
    ):

        # add convolutional layer
        conv_layer_name = "conv_" + str(conv_layer_number + 1)
        if conv_layer_number == 0:
            conv_layer = layers.Conv2D(
                conv_layer_dimension,
                (3, 3),
                activation="relu",
                padding="same",
                name=conv_layer_name,
            )
        else:
            conv_layer = layers.Conv2D(
                conv_layer_dimension,
                (3, 3),
                activation="relu",
                padding="same",
                name=conv_layer_name,
            )
        if conv_layer_number == 0:
            network.add(layers.TimeDistributed(conv_layer, input_shape=input_shape))
        else:
            network.add(layers.TimeDistributed(conv_layer))

        # add pooling layer
        pooling_layer_name = "pooling_" + str(conv_layer_number + 1)
        pooling_layer = layers.MaxPooling2D(2, 2, name=pooling_layer_name)
        network.add(layers.TimeDistributed(pooling_layer))
    # FLATTENING
    flatten_layer_name = "flatten"
    flatten_layer = layers.Flatten(name=flatten_layer_name)
    network.add(layers.TimeDistributed(flatten_layer))

    # DENSE TOP
    for dense_layer_number, dense_layer_dimension in zip(
        range(len(dense_layers_dimensions)), dense_layers_dimensions
    ):

        # add dense layer
        dense_layer_name = "dense_" + str(dense_layer_number + 1)
        dense_layer = layers.Dense(
            dense_layer_dimension, activation="relu", name=dense_layer_name
        )
        network.add(layers.TimeDistributed(dense_layer))

    for rnn_layer_number, rnn_layer_dimension in zip(
        range(len(rnn_layers_dimensions)), rnn_layers_dimensions
    ):

        # add dense layer
        rnn_layer_name = "rnn_" + str(rnn_layer_number + 1)
        rnn_layer = layers.LSTM(
            rnn_layer_dimension,
            name=rnn_layer_name,
            return_sequences=rnn_layer_number < len(rnn_layers_dimensions) - 1
            or return_sequences,
        )
        network.add(rnn_layer)

    # OUTPUT LAYER

    output_layer = layers.Dense(
        number_of_outputs, activation=output_activation, name="output"
    )
    if return_sequences:
        network.add(layers.TimeDistributed(output_layer))
    else:
        network.add(output_layer)

    return KerasModel(network, **kwargs)


# Alias for backwards compatability
rnn = RNN


class cgan(Model):
    def __init__(
        self,
        generator=None,
        discriminator=None,
        discriminator_loss=None,
        discriminator_optimizer=None,
        discriminator_metrics=None,
        assemble_loss=None,
        assemble_optimizer=None,
        assemble_loss_weights=None,
        **kwargs
    ):

        # Build and compile the discriminator
        self.discriminator = discriminator
        self.discriminator.compile(
            loss=discriminator_loss,
            optimizer=discriminator_optimizer,
            metrics=discriminator_metrics,
        )

        # Build the generator
        self.generator = generator

        # Input shape
        self.model_input = self.generator.input

        # The generator model_input and generates img
        img = self.generator(self.model_input)

        # For the combined model we will only train the generator
        self.discriminator.trainable = False

        # The discriminator takes the generated images as input and determines validity
        validity = self.discriminator([img, self.model_input])

        # The assembled model (stacked generator and discriminator)
        # Trains the generator to fool the discriminator
        self.assemble = models.Model(self.model_input, [validity, img])
        self.assemble.compile(
            loss=assemble_loss,
            optimizer=assemble_optimizer,
            loss_weights=assemble_loss_weights,
        )

        super().__init__(self.generator, **kwargs)

    def fit(self, data_generator, epochs, steps_per_epoch=None, **kwargs):
        for key in kwargs.keys():
            warnings.warn(
                "{0} not implemented for cgan. Does not affect the execution.".format(
                    key
                )
            )
        for epoch in range(epochs):
            if not steps_per_epoch:
                try:
                    steps = len(data_generator)
                except:
                    steps = 1
            else:
                steps = steps_per_epoch

            d_loss = 0
            g_loss = 0

            for step in range(steps):
                # update data
                try:
                    data, labels = next(data_generator)
                except:
                    data, labels = data_generator[step]

                # Grab disriminator labels
                shape = (data.shape[0], *self.discriminator.output.shape[1:])
                valid, fake = np.ones(shape), np.zeros(shape)

                # ---------------------
                #  Train Discriminator
                # ---------------------

                # Generate a batch of new images
                gen_imgs = self.generator(data)

                # Train the discriminator
                d_loss_real = self.discriminator.train_on_batch([labels, data], valid)
                d_loss_fake = self.discriminator.train_on_batch([gen_imgs, data], fake)
                d_loss += 0.5 * np.add(d_loss_real, d_loss_fake)

                # ---------------------
                #  Train Generator
                # ---------------------

                # Train the generator (to have the discriminator label samples as valid)
                g_loss += np.array(self.assemble.train_on_batch(data, [valid, labels]))

                # Plot the progress

            try:
                data_generator.on_epoch_end()
            except:
                pass

            print(
                "%d [D loss: %f, acc.: %.2f%%] [G loss: %f, %f, %f]"
                % (
                    epoch,
                    d_loss[0] / steps,
                    100 * d_loss[1] / steps,
                    g_loss[0] / steps,
                    g_loss[1] / steps,
                    g_loss[2] / steps,
                )
            )



import sys

sys.path.append("..")  # Adds the module to path

import unittest

import deeptrack.layers as layers
import tensorflow.keras.layers as k_layers
import tensorflow.keras.models as k_models
from tensorflow_addons.layers import InstanceNormalization

import numpy as np


def makeMinimalModel(layer, shape) -> k_models.Model:

    i = k_layers.Input(shape=shape)
    o = layer(i)

    return k_models.Model(i, o)


class TestModels(unittest.TestCase):
    def test_Dense(self):
        block = layers.DenseBlock()
        model = makeMinimalModel(block(1), shape=(2,))
        self.assertIsInstance(model.layers[1], k_layers.Dense)

    def test_Conv(self):
        block = layers.ConvolutionalBlock()
        model = makeMinimalModel(block(1), shape=(4, 4, 1))
        self.assertIsInstance(model.layers[1], k_layers.Conv2D)

    def test_Pool(self):
        block = layers.PoolingBlock()
        model = makeMinimalModel(block(1), shape=(2, 2, 1))
        self.assertIsInstance(model.layers[1], k_layers.MaxPool2D)

    def test_Deconv(self):
        block = layers.DeconvolutionalBlock()
        model = makeMinimalModel(block(1), shape=(2, 2, 1))
        self.assertIsInstance(model.layers[1], k_layers.Conv2DTranspose)

    def test_StaticUpsample(self):
        block = layers.StaticUpsampleBlock()
        model = makeMinimalModel(block(1), shape=(2, 2, 1))
        self.assertIsInstance(model.layers[1], k_layers.UpSampling2D)

    def test_Residual(self):
        block = layers.ResidualBlock()
        model = makeMinimalModel(block(1), shape=(2, 2, 1))
        self.assertEqual(len(model.layers), 9)

    def test_Identity(self):
        block = layers.Identity()
        model = makeMinimalModel(block(1), shape=(2, 2, 1))
        self.assertIsInstance(model.layers[1], k_layers.Layer)

    def test_NoneActivation(self):
        block = layers.Identity(activation=None)
        model = makeMinimalModel(block(1), shape=(2, 2, 1))
        self.assertEqual(len(model.layers), 2)

    def test_InstanceNorm(self):
        block = layers.Identity(instance_norm=True)
        model = makeMinimalModel(block(1), shape=(2, 2, 1))
        self.assertIsInstance(model.layers[2], InstanceNormalization)


if __name__ == "__main__":
    unittest.main()


import sys

sys.path.append("..")  # Adds the module to path

import unittest

import deeptrack.losses as losses

import numpy as np
from keras import backend as K


class TestLosses(unittest.TestCase):

    truthly = K.constant(np.ones((2, 2, 1)))
    falsely = K.constant(np.zeros((2, 2, 1)))

    def test_flatten(self):
        from keras.losses import mse

        loss_function = losses.flatten(mse)
        loss = K.eval(loss_function(self.truthly, self.truthly))
        self.assertAlmostEqual(loss, 0, 3)
        loss = K.eval(loss_function(self.truthly, self.falsely))
        self.assertAlmostEqual(loss, 1, 3)
        loss = K.eval(loss_function(self.falsely, self.truthly))
        self.assertAlmostEqual(loss, 1, 3)
        loss = K.eval(loss_function(self.falsely, self.falsely))
        self.assertAlmostEqual(loss, 0, 3)

    def test_weighted_crossentropy(self):
        loss_function = losses.flatten(losses.weighted_crossentropy(weight=(10, 1)))
        loss = K.eval(loss_function(self.truthly, self.truthly))
        self.assertAlmostEqual(loss, 0, 3)
        loss = K.eval(loss_function(self.truthly, self.falsely))
        self.assertAlmostEqual(loss, 8.373037, 3)
        loss = K.eval(loss_function(self.falsely, self.truthly))
        self.assertAlmostEqual(loss, 0.8373037, 3)
        loss = K.eval(loss_function(self.falsely, self.falsely))
        self.assertAlmostEqual(loss, 0, 3)

    def test_nd_mean_squared_error(self):
        loss_function = losses.nd_mean_squared_error
        loss = K.eval(loss_function(self.truthly, self.truthly))
        self.assertAlmostEqual(loss, 0, 3)
        loss = K.eval(loss_function(self.truthly, self.falsely))
        self.assertAlmostEqual(loss, 1, 3)
        loss = K.eval(loss_function(self.falsely, self.truthly))
        self.assertAlmostEqual(loss, 1, 3)
        loss = K.eval(loss_function(self.falsely, self.falsely))
        self.assertAlmostEqual(loss, 0, 3)

    def test_nd_mean_squared_logarithmic_error(self):
        loss_function = losses.nd_mean_squared_logarithmic_error
        loss = K.eval(loss_function(self.truthly, self.truthly))
        self.assertAlmostEqual(loss, 0, 3)
        loss = K.eval(loss_function(self.truthly, self.falsely))
        self.assertAlmostEqual(loss, 0.48045287, 3)
        loss = K.eval(loss_function(self.falsely, self.truthly))
        self.assertAlmostEqual(loss, 0.48045287, 3)
        loss = K.eval(loss_function(self.falsely, self.falsely))
        self.assertAlmostEqual(loss, 0, 3)

    def test_nd_poisson(self):
        loss_function = losses.nd_poisson
        loss = K.eval(loss_function(self.truthly, self.truthly))
        self.assertAlmostEqual(loss, 0.9999999, 3)
        loss = K.eval(loss_function(self.truthly, self.falsely))
        self.assertAlmostEqual(loss, 16.118095, 3)
        loss = K.eval(loss_function(self.falsely, self.truthly))
        self.assertAlmostEqual(loss, 1, 3)
        loss = K.eval(loss_function(self.falsely, self.falsely))
        self.assertAlmostEqual(loss, 0, 3)

    def test_nd_squared_hinge(self):
        loss_function = losses.nd_squared_hinge
        loss = K.eval(loss_function(self.truthly, self.truthly))
        self.assertAlmostEqual(loss, 0, 3)
        loss = K.eval(loss_function(self.truthly, self.falsely))
        self.assertAlmostEqual(loss, 1, 3)
        loss = K.eval(loss_function(self.falsely, self.truthly))
        self.assertAlmostEqual(loss, 4, 3)
        loss = K.eval(loss_function(self.falsely, self.falsely))
        self.assertAlmostEqual(loss, 1, 3)

    def test_nd_binary_crossentropy(self):
        loss_function = losses.nd_binary_crossentropy
        loss = K.eval(loss_function(self.truthly, self.truthly))
        self.assertAlmostEqual(loss, 0, 3)
        loss = K.eval(loss_function(self.truthly, self.falsely))
        self.assertAlmostEqual(loss, 15.424949, 3)
        loss = K.eval(loss_function(self.falsely, self.truthly))
        self.assertAlmostEqual(loss, 15.333239, 3)
        loss = K.eval(loss_function(self.falsely, self.falsely))
        self.assertAlmostEqual(loss, 0, 3)

    def test_nd_mean_absolute_percentage_error(self):
        loss_function = losses.nd_mean_absolute_percentage_error
        loss = K.eval(loss_function(self.truthly, self.truthly))
        self.assertAlmostEqual(loss, 0, 3)
        loss = K.eval(loss_function(self.truthly, self.falsely))
        self.assertAlmostEqual(loss, 100, 3)
        loss = K.eval(loss_function(self.falsely, self.truthly))
        self.assertAlmostEqual(loss, 1000000000, 3)
        loss = K.eval(loss_function(self.falsely, self.falsely))
        self.assertAlmostEqual(loss, 0, 3)


if __name__ == "__main__":
    unittest.main()


""" Mathematical oprations and structures

Classses
--------
Clip
    Clip the input within a minimum and a maximum value.
NormalizeMinMax
    Min-max image normalization.
"""

from deeptrack.features import Feature
from deeptrack.image import Image
import numpy as np


class Add(Feature):
    """Adds a value to the input.

    Parameters
    ----------
    value : number
        The value to add
    """

    def __init__(self, value=0, **kwargs):
        super().__init__(value=value, **kwargs)

    def get(self, image, value, **kwargs):
        return image + value


class Subtract(Feature):
    """Subtracts a value from the input.

    Parameters
    ----------
    value : number
        The value to subtract
    """

    def __init__(self, value=0, **kwargs):
        super().__init__(value=value, **kwargs)

    def get(self, image, value, **kwargs):
        return image - value


class Multiply(Feature):
    """Multiplies the input with a value.

    Parameters
    ----------
    value : number
        The value to multiply with.
    """

    def __init__(self, value=0, **kwargs):
        super().__init__(value=value, **kwargs)

    def get(self, image, value, **kwargs):
        return image * value


class Divide(Feature):
    """Divides the input with a value.

    Parameters
    ----------
    value : number
        The value to divide with.
    """

    def __init__(self, value=0, **kwargs):
        super().__init__(value=value, **kwargs)

    def get(self, image, value, **kwargs):
        return image / value


class Power(Feature):
    """Raises the input to a power.

    Parameters
    ----------
    value : number
        The power to raise with.
    """

    def __init__(self, value=0, **kwargs):
        super().__init__(value=value, **kwargs)

    def get(self, image, value, **kwargs):
        return image ** value


class Average(Feature):
    """Average of input images

    If `features` is not None, it instead resolves all features
    in the list and averages the result.

    Parameters
    ----------
    axis : int or tuple of ints
        Axis along which to average
    features : list of features, optional
    """

    __distributed__ = False

    def __init__(self, features=None, axis=0, **kwargs):
        super().__init__(axis=axis, features=features, **kwargs)

    def get(self, images, axis, features, **kwargs):
        if features is not None:
            images = [feature.resolve() for feature in features]
        result = Image(np.mean(images, axis=axis))

        for image in images:
            result.merge_properties_from(image)

        return result


class Clip(Feature):
    """Clip the input within a minimum and a maximum value.

    Parameters
    ----------
    min : float
        Clip the input to be larger than this value.
    max : float
        Clip the input to be smaller than this value.
    """

    def __init__(self, min=-np.inf, max=+np.inf, **kwargs):
        super().__init__(min=min, max=max, **kwargs)

    def get(self, image, min=None, max=None, **kwargs):
        np.clip(image, min, max, image)
        return image


class NormalizeMinMax(Feature):
    """Image normalization.

    Transforms the input to be between a minimum and a maximum value using a linear transformation.

    Parameters
    ----------
    min : float
        The minimum of the transformation.
    max : float
        The maximum of the transformation.
    """

    def __init__(self, min=0, max=1, **kwargs):
        super().__init__(min=min, max=max, **kwargs)

    def get(self, image, min, max, **kwargs):
        image = image / (np.max(image) - np.min(image)) * (max - min)
        image = image - np.min(image) + min
        image[np.isnan(image)] = 0
        return image


import deeptrack.utils as utils
import skimage

import scipy.ndimage as ndimage


class Blur(Feature):
    def __init__(self, filter_function, mode="reflect", **kwargs):
        self.filter = filter_function
        super().__init__(borderType=borderType, **kwargs)

    def get(self, image, **kwargs):
        kwargs.pop("input", False)
        return utils.safe_call(self.filter, input=image, **kwargs)


class AverageBlur(Blur):
    """Blur an image by computing simple means over neighbourhoods.

    Performs a (N-1)D convolution if the last dimension is smaller than the kernel size.

    Parameters
    ----------
    ksize : int
        Kernel size to use.
    """

    def __init__(self, ksize=3, **kwargs):
        super().__init__(None, ksize=ksize, **kwargs)

    def get(self, input, ksize, **kwargs):

        if input.shape[-1] < ksize:
            ksize = (ksize,) * (input.ndim - 1) + (1,)
        else:
            ksize = ((ksize,) * input.ndim,)

        weights = np.ones(ksize) / np.prod(ksize)

        return safe_call(ndimage, input=input, weights=weights, **kwargs)


class GaussianBlur(Blur):
    """Augmenter to blur images using gaussian kernels.

    Parameters
    ----------
    sigma : number
        Standard deviation of the gaussian kernel.

    """

    def __init__(self, sigma=2, **kwargs):
        super().__init__(ndimage.gaussian_filter, sigma=sigma, **kwargs)


class MedianBlur(Blur):
    """Blur an image by computing median values over neighbourhoods.

    Parameters
    ----------
    ksize :
        Kernel size.

    """

    def __init__(self, ksize=3, **kwargs):
        super().__init__(ndimage.median_filter, k=k, **kwargs)


## POOLING

import skimage


class Pool(Feature):
    """Downsamples the image by applying a function to local regions of the image.

    Parameters
    ----------
    pooling_function : function
        A function that is applied to each local region of the image.
        DOES NOT NEED TO BE WRAPPED IN A ANOTHER FUNCTIOn.
        Must implement the axis argument. Examples include
        np.mean, np.max, np.min, etc.
    ksize : int
        Size of the pooling kernel.
    cval : number
        Value to pad edges with if necessary.
    func_kwargs : dict
        Additional parameters sent to the pooling function.
    """

    def __init__(self, pooling_function, ksize=2, **kwargs):
        self.pooling = pooling_function
        super().__init__(ksize=ksize, **kwargs)

    def get(self, image, ksize, **kwargs):
        kwargs.pop("func", False)
        kwargs.pop("image", False)
        kwargs.pop("block_size", False)
        return utils.safe_call(
            skimage.measure.block_reduce,
            image=image,
            func=self.pooling,
            block_size=ksize,
            **kwargs
        )


class AveragePooling(Pool):
    """Apply average pooling to images.

    Parameters
    ----------
    ksize : int
        Size of the pooling kernel.
    cval : number
        Value to pad edges with if necessary. Default 0.
    func_kwargs : dict
        Additional parameters sent to the pooling function.
    """

    def __init__(self, ksize=3, **kwargs):
        super().__init__(ksize=ksize, **kwargs)


### OPENCV2 blur

try:
    import cv2

    IMPORTED_CV2 = True

    _map_mode_to_cv2_borderType = {
        "reflect": cv2.BORDER_REFLECT,
        "wrap": cv2.BORDER_WRAP,
        "constant": cv2.BORDER_CONSTANT,
        "mirror": cv2.BORDER_REFLECT_101,
        "nearest": cv2.BORDER_REPLICATE,
    }
except ImportError:
    IMPORTED_CV2 = False


class BlurCV2(Feature):
    def __new__(cls, *args, **kwargs):
        if not IMPORTED_CV2:
            raise ImportError(
                "opencv not installed on device, it is an optional dependency of deeptrack. To use this feature, you need to install it manually."
            )

        return super(BlurCV2, cls).__new__(*args, **kwargs)

    def __init__(self, filter_function, mode="refelct", **kwargs):
        self.filter = filter_function
        borderType = _map_mode_to_cv2_borderType[mode]
        super().__init__(borderType=borderType, **kwargs)

    def get(self, image, **kwargs):
        kwargs.pop("src", False)
        kwargs.pop("dst", False)
        utils.safe_call(self.filter, src=image, dst=image, **kwargs)
        return image


class BilateralBlur(Blur):
    """Blur an image using a bilateral filter.

    Bilateral filters blur homogenous areas while trying to
    preserve edges.


    Parameters
    ----------
    d : int
        Diameter of each pixel neighborhood with value range.

    sigma_color : number
        Filter sigma in the color space with value range. A
        large value of the parameter means that farther colors within the
        pixel neighborhood (see `sigma_space`) will be mixed together,
        resulting in larger areas of semi-equal color.

    sigma_space : number
        Filter sigma in the coordinate space with value range. A
        large value of the parameter means that farther pixels will influence
        each other as long as their colors are close enough (see
        `sigma_color`).

    """

    def __init__(self, d=3, sigma_color=50, sigma_space=50, **kwargs):
        super().__init__(
            cv2.bilateralFilter,
            d=d,
            sigma_color=sigma_color,
            sigma_space=sigma_space,
            **kwargs
        )


import sys

sys.path.append("..")  # Adds the module to path

import unittest

import deeptrack.utils as utils


class TestUtils(unittest.TestCase):
    def test_hasmethod(self):
        self.assertTrue(utils.hasmethod(utils, "hasmethod"))
        self.assertFalse(
            utils.hasmethod(utils, "this_is_definetely_not_a_method_of_utils")
        )

    def test_isiterable(self):
        self.assertFalse(utils.isiterable(1))

        non_iterable_obj = ("apple", "banana", "cherry")
        self.assertFalse(utils.isiterable(non_iterable_obj))

        iterable_obj = iter(("apple", "banana", "cherry"))
        self.assertTrue(utils.isiterable(iterable_obj))

    def test_as_list(self):

        obj = 1
        self.assertEqual(utils.as_list(obj), [obj])

        list_obj = [1, 2, 3]
        self.assertEqual(utils.as_list(list_obj), list_obj)

    def test_get_kwarg_names(self):
        def func1():
            pass

        self.assertEqual(utils.get_kwarg_names(func1), [])

        def func2(key1):
            pass

        self.assertEqual(utils.get_kwarg_names(func2), ["key1"])

        def func3(key1, key2=2):
            pass

        self.assertEqual(utils.get_kwarg_names(func3), ["key1", "key2"])

        def func4(key1, *argv, key2=2):
            pass

        self.assertEqual(utils.get_kwarg_names(func4), ["key2"])

        def func5(*argv, key1, key2=2):
            pass

        self.assertEqual(utils.get_kwarg_names(func5), ["key1", "key2"])

        def func6(key1, key2, key3, *argv):
            pass

        self.assertEqual(utils.get_kwarg_names(func6), [])

        def func7(key1, key2=1, key3=3, **kwargs):
            pass

        self.assertEqual(utils.get_kwarg_names(func7), ["key1", "key2", "key3"])

    def test_safe_call(self):

        arguments = {
            "key1": None,
            "key2": False,
            "key_not_in_function": True,
            "key_not_in_function_2": True,
        }

        def func1():
            pass

        utils.safe_call(func1, **arguments)

        def func2(key1):
            pass

        utils.safe_call(func2, **arguments)

        def func3(key1, key2=2):
            pass

        utils.safe_call(func3, **arguments)

        def func4(key1, *argv, key2=2):
            pass

        self.assertRaises(TypeError, lambda: utils.safe_call(func4, **arguments))

        def func5(*argv, key1, key2=2):
            pass

        utils.safe_call(func5, **arguments)

        def func6(key1, key2=1, key3=3, **kwargs):
            pass

        utils.safe_call(func6, **arguments)


if __name__ == "__main__":
    unittest.main()


import sys

sys.path.append("..")  # Adds the module to path

import unittest

import deeptrack.generators as generators

from deeptrack.optics import Fluorescence
from deeptrack.scatterers import PointParticle
import numpy as np


class TestGenerators(unittest.TestCase):
    def test_Generator(self):
        optics = Fluorescence(
            NA=0.7,
            wavelength=680e-9,
            resolution=1e-6,
            magnification=10,
            output_region=(0, 0, 128, 128),
        )
        scatterer = PointParticle(
            intensity=100,
            position_unit="pixel",
            position=lambda: np.random.rand(2) * 128,
        )
        imaged_scatterer = optics(scatterer)

        def get_particle_position(result):
            for property in result.properties:
                if "position" in property:
                    return property["position"]

        generator = generators.Generator()
        particle_generator = generator.generate(
            imaged_scatterer, get_particle_position, batch_size=4
        )
        particles, positions = next(particle_generator)
        for particle, position in zip(particles, positions):
            self.assertEqual(particle.shape, (128, 128, 1))
            self.assertTrue(np.all(position >= 0))
            self.assertTrue(np.all(position <= 128))


if __name__ == "__main__":
    unittest.main()


import sys

sys.path.append("..")  # Adds the module to path

import unittest

import deeptrack.noises as noises

from deeptrack.image import Image
import numpy as np


class TestNoises(unittest.TestCase):
    def test_Offset(self):
        noise = noises.Offset(offset=0.5)
        input_image = Image(np.zeros((256, 256)))
        output_image = noise.resolve(input_image)
        self.assertIsInstance(output_image, Image)
        self.assertEqual(output_image.shape, (256, 256))
        self.assertTrue(np.all(np.array(output_image) == 0.5))

    def test_Background(self):
        noise = noises.Background(offset=0.5)
        input_image = Image(np.zeros((256, 256)))
        output_image = noise.resolve(input_image)
        self.assertIsInstance(output_image, Image)
        self.assertEqual(output_image.shape, (256, 256))
        self.assertTrue(np.all(np.array(output_image) == 0.5))

    def test_Gaussian(self):
        noise = noises.Gaussian(mu=0.1, sigma=0.05)
        input_image = Image(np.zeros((256, 256)))
        output_image = noise.resolve(input_image)
        self.assertIsInstance(output_image, Image)
        self.assertEqual(output_image.shape, (256, 256))

    def test_Poisson(self):
        noise = noises.Poisson(snr=20)
        input_image = Image(np.ones((256, 256)) * 0.1)
        output_image = noise.resolve(input_image)
        self.assertIsInstance(output_image, Image)
        self.assertEqual(output_image.shape, (256, 256))


if __name__ == "__main__":
    unittest.main()


"""Implementations of Feature the model scattering objects.

Provides some basic implementations of scattering objects 
that are frequently used.

Classes
--------
Scatterer
    Abstract base class for scatterers
PointParticle
    Generates point particles
Ellipse
    Generetes 2-d elliptical particles
Sphere
    Generates 3-d spheres
Ellipsoid
    Generates 3-d ellipsoids
"""

from threading import Lock
import numpy as np
from scipy.special import jv as jn, spherical_jn as jv, h1vp, eval_legendre as leg, jvp
from deeptrack.features import Feature, MERGE_STRATEGY_APPEND
from deeptrack.image import Image
import deeptrack.image


class Scatterer(Feature):
    """Base abstract class for scatterers.

    A scatterer is defined by a 3-dimensional volume of voxels.
    To each voxel corresponds an occupancy factor, i.e., how much
    of that voxel does the scatterer occupy. However, this number is not
    necessarily limited to the [0, 1] range. It can be any number, and its
    interpretation is left to the optical device that images the scatterer.

    This abstract class implements the `_process_properties` method to convert
    the position to voxel units, as well as the `_process_and_get` method to
    upsample the calculation and crop empty slices.

    Parameters
    ----------
    position : array_like of length 2 or 3
        The position of the  particle. Third index is optional,
        and represents the position in the direction normal to the
        camera plane.
    z : float
        The position in the direction normal to the
        camera plane. Used if `position` is of length 2.
    value : float
        A default value of the characteristic of the particle. Used by
        optics unless a more direct property is set (eg. `refractive_index`
        for `Brightfield` and `intensity` for `Fluorescence`).
    position_unit : "meter" or "pixel"
        The unit of the provided position property.

    Other Parameters
    ----------------
    upsample_axes : tuple of ints
        Sets the axes along which the calculation is upsampled (default is None,
        which implies all axes are upsampled).
    crop_zeros : bool
        Whether to remove slices in which all elements are zero.
    """

    __list_merge_strategy__ = MERGE_STRATEGY_APPEND
    __distributed__ = False

    def __init__(
        self,
        position=(32, 32),
        z=0.0,
        value=1.0,
        position_unit="pixel",
        upsample=1,
        **kwargs
    ):
        self._processed_properties = False
        super().__init__(
            position=position,
            z=z,
            value=value,
            position_unit=position_unit,
            upsample=upsample,
            **kwargs
        )

    def _process_properties(self, properties: dict) -> dict:
        # Rescales the position property
        self._processed_properties = True
        if "position" in properties:
            if properties["position_unit"] == "meter":
                properties["position"] = (
                    np.array(properties["position"])
                    / np.array(properties["voxel_size"])[: len(properties["position"])]
                    / properties.get("upscale", 1)
                )
                properties["z"] = (
                    np.array(properties["z"])
                    / np.array(properties["voxel_size"])[: len(properties["position"])]
                    / properties.get("upscale", 1)
                )

        return properties

    def _process_and_get(
        self, *args, voxel_size, upsample, upsample_axes=None, crop_empty=True, **kwargs
    ):
        # Post processes the created object to handle upsampling,
        # as well as cropping empty slices.
        if not self._processed_properties:
            import warnings

            warnings.warn(
                "Overridden _process_properties method does not call super. This is likely to result in errors if used with Optics.upscale != 1."
            )

        # Calculates upsampled voxel_size
        if upsample_axes is None:
            upsample_axes = range(3)

        voxel_size = np.array(voxel_size)
        for axis in upsample_axes:
            voxel_size[axis] /= upsample

        # calls parent _process_and_get
        new_image = super()._process_and_get(
            *args, voxel_size=voxel_size, upsample=upsample, **kwargs
        )
        new_image = new_image[0]

        # Downsamples the image along the axes it was upsampled
        if upsample != 1 and upsample_axes:

            # Pad image to ensure it is divisible by upsample
            increase = np.array(new_image.shape)
            for axis in upsample_axes:
                increase[axis] = upsample - (new_image.shape[axis] % upsample)
            pad_width = [(0, inc) for inc in increase]
            new_image = np.pad(new_image, pad_width, mode="constant")

            # Finds reshape size for downsampling
            new_shape = []
            for axis in range(new_image.ndim):
                if axis in upsample_axes:
                    new_shape += [new_image.shape[axis] // upsample, upsample]
                else:
                    new_shape += [new_image.shape[axis]]

            # Downsamples
            new_image = np.reshape(new_image, new_shape).mean(
                axis=tuple(np.array(upsample_axes, dtype=np.int32) * 2 + 1)
            )

        # Crops empty slices
        if crop_empty:
            new_image = new_image[~np.all(new_image == 0, axis=(1, 2))]
            new_image = new_image[:, ~np.all(new_image == 0, axis=(0, 2))]
            new_image = new_image[:, :, ~np.all(new_image == 0, axis=(0, 1))]

        return [Image(new_image)]


class PointParticle(Scatterer):
    """Generates a point particle

    A point particle is approximated by the size of a pixel. For subpixel
    positioning, the position is interpolated linearly.

    Parameters
    ----------
    position : array_like of length 2 or 3
        The position of the  particle. Third index is optional,
        and represents the position in the direction normal to the
        camera plane.
    z : float
        The position in the direction normal to the
        camera plane. Used if `position` is of length 2.
    value : float
        A default value of the characteristic of the particle. Used by
        optics unless a more direct property is set: (eg. `refractive_index`
        for `Brightfield` and `intensity` for `Fluorescence`).
    """

    def __init__(self, **kwargs):
        kwargs.pop("upsample", False)
        kwargs.pop("upsample_axes", False)
        super().__init__(upsample=1, upsample_axes=(), **kwargs)

    def get(self, image, **kwargs):
        return np.ones((1, 1, 1))


class Ellipse(Scatterer):
    """Generates an elliptical disk scatterer

    Parameters
    ----------
    radius : float or array_like [float (, float)]
        Radius of the ellipse in meters. If only one value,
        assume circular.
    rotation : float
        Orientation angle of the ellipse in the camera plane in radians.
    position : array_like[float, float (, float)]
        The position of the particle. Third index is optional,
        and represents the position in the direction normal to the
        camera plane.
    z : float
        The position in the direction normal to the
        camera plane. Used if `position` is of length 2.
    value : float
        A default value of the characteristic of the particle. Used by
        optics unless a more direct property is set: (eg. `refractive_index`
        for `Brightfield` and `intensity` for `Fluorescence`).
    upsample : int
        Upsamples the calculations of the pixel occupancy fraction.
    """

    def __init__(self, radius=1e-6, rotation=0, **kwargs):
        super().__init__(
            radius=radius, rotation=rotation, upsample_axes=(0, 1), **kwargs
        )

    def _process_properties(self, properties: dict) -> dict:
        """Preprocess the input to the method .get()

        Ensures that the radius is an array of length 2. If the radius
        is a single value, the particle is made circular
        """

        properties = super()._process_properties(properties)

        # Ensure radius is of length 2
        radius = np.array(properties["radius"])
        if radius.ndim == 0:
            radius = np.array((properties["radius"], properties["radius"]))
        elif radius.size == 1:
            radius = np.array((*radius,) * 2)
        else:
            radius = radius[:2]
        properties["radius"] = radius

        return properties

    def get(self, *ignore, radius, rotation, voxel_size, **kwargs):

        # Create a grid to calculate on
        rad = radius[:2] / voxel_size[:2]
        ceil = int(np.max(np.ceil(rad)))
        X, Y = np.meshgrid(np.arange(-ceil, ceil), np.arange(-ceil, ceil))

        # Rotate the grid
        if rotation != 0:
            Xt = X * np.cos(-rotation) + Y * np.sin(-rotation)
            Yt = -X * np.sin(-rotation) + Y * np.cos(-rotation)
            X = Xt
            Y = Yt

        # Evaluate ellipse
        mask = ((X * X) / (rad[0] * rad[0]) + (Y * Y) / (rad[1] * rad[1]) < 1) * 1.0
        mask = np.expand_dims(mask, axis=-1)
        return mask


class Sphere(Scatterer):
    """Generates a spherical scatterer

    Parameters
    ----------
    radius : float
        Radius of the sphere in meters.
    position : array_like[float, float (, float)]
        The position of the particle. Third index is optional,
        and represents the position in the direction normal to the
        camera plane.
    z : float
        The position in the direction normal to the
        camera plane. Used if `position` is of length 2.
    value : float
        A default value of the characteristic of the particle. Used by
        optics unless a more direct property is set: (eg. `refractive_index`
        for `Brightfield` and `intensity` for `Fluorescence`).
    upsample : int
        Upsamples the calculations of the pixel occupancy fraction.
    """

    def __init__(self, radius=1e-6, **kwargs):
        super().__init__(radius=radius, **kwargs)

    def get(self, image, radius, voxel_size, **kwargs):

        # Create a grid to calculate on
        rad = radius / voxel_size
        rad_ceil = np.ceil(rad)
        x = np.arange(-rad_ceil[0], rad_ceil[0])
        y = np.arange(-rad_ceil[1], rad_ceil[1])
        z = np.arange(-rad_ceil[2], rad_ceil[2])
        X, Y, Z = np.meshgrid((x / rad[0]) ** 2, (y / rad[1]) ** 2, (z / rad[2]) ** 2)

        mask = (X + Y + Z <= 1) * 1.0

        return mask


class Ellipsoid(Scatterer):
    """Generates an ellipsoidal scatterer

    Parameters
    ----------
    radius : float or array_like[float (, float, float)]
        Radius of the ellipsoid in meters. If only one value,
        assume spherical.
    rotation : float
        Rotation of the ellipsoid in about the x, y and z axis.
    position : array_like[float, float (, float)]
        The position of the particle. Third index is optional,
        and represents the position in the direction normal to the
        camera plane.
    z : float
        The position in the direction normal to the
        camera plane. Used if `position` is of length 2.
    value : float
        A default value of the characteristic of the particle. Used by
        optics unless a more direct property is set: (eg. `refractive_index`
        for `Brightfield` and `intensity` for `Fluorescence`).
    upsample : int
        Upsamples the calculations of the pixel occupancy fraction.
    """

    def __init__(self, radius=1e-6, rotation=0, **kwargs):
        super().__init__(radius=radius, rotation=rotation, **kwargs)

    def _process_properties(self, propertydict):
        """Preprocess the input to the method .get()

        Ensures that the radius and the rotation properties both are arrays of
        length 3.

        If the radius is a single value, the particle is made a sphere
        If the radius are two values, the smallest value is appended as the third value

        The rotation vector is padded with zeros until it is of length 3
        """

        propertydict = super()._process_properties(propertydict)

        # Ensure radius has three values
        radius = np.array(propertydict["radius"])
        if radius.ndim == 0:
            radius = np.array([radius])
        if radius.size == 1:
            # If only one value, assume sphere
            radius = (*radius,) * 3
        elif radius.size == 2:
            # If two values, duplicate the minor axis
            radius = (*radius, np.min(radius[-1]))
        elif radius.size == 3:
            # If three values, convert to tuple for consistency
            radius = (*radius,)
        propertydict["radius"] = radius

        # Ensure rotation has three values
        rotation = np.array(propertydict["rotation"])
        if rotation.ndim == 0:
            rotation = np.array([rotation])
        if rotation.size == 1:
            # If only one value, pad with two zeros
            rotation = (*rotation, 0, 0)
        elif rotation.size == 2:
            # If two values, pad with one zero
            rotation = (*rotation, 0)
        elif rotation.size == 3:
            # If three values, convert to tuple for consistency
            rotation = (*rotation,)
        propertydict["rotation"] = rotation

        return propertydict

    def get(self, image, radius, rotation, voxel_size, **kwargs):

        radius_in_pixels = radius / voxel_size

        max_rad = np.max(radius) / voxel_size
        rad_ceil = np.ceil(max_rad)

        # Create grid to calculate on
        x = np.arange(-rad_ceil[0], rad_ceil[0])
        y = np.arange(-rad_ceil[1], rad_ceil[1])
        z = np.arange(-rad_ceil[2], rad_ceil[2])
        X, Y, Z = np.meshgrid(x, y, z)

        # Rotate the grid
        cos = np.cos(rotation)
        sin = np.sin(rotation)
        XR = (
            (cos[0] * cos[1] * X)
            + (cos[0] * sin[1] * sin[2] - sin[0] * cos[2]) * Y
            + (cos[0] * sin[1] * cos[2] + sin[0] * sin[2]) * Z
        )
        YR = (
            (sin[0] * cos[1] * X)
            + (sin[0] * sin[1] * sin[2] + cos[0] * cos[2]) * Y
            + (sin[0] * sin[1] * cos[2] - cos[0] * sin[2]) * Z
        )
        ZR = (-sin[1] * X) + cos[1] * sin[2] * Y + cos[1] * cos[2] * Z

        mask = (
            (XR / radius_in_pixels[0]) ** 2
            + (YR / radius_in_pixels[1]) ** 2
            + (ZR / radius_in_pixels[2]) ** 2
            < 1
        ) * 1.0

        return mask


class MieScatterer(Scatterer):
    def __init__(self, **kwargs):
        kwargs.pop("is_field", None)
        kwargs.pop("crop_empty", None)
        super().__init__(is_field=True, crop_empty=False, **kwargs)


# class MieBisphere(MieScatterer):
#     def __init__(self, refractive_index=1.45, distance=2e-6, radius=1e-6, **kwargs):
#         super().__init__(
#             refractive_index=refractive_index,
#             distance=distance,
#             radius=radius,
#             **kwargs
#         )


class MieSphere(MieScatterer):
    """Scattered field by a sphere

    Calculates the scattered field by a spherical particle in a homogenous medium,
    as predicted by Mie theory. Note that the induced phase shift is calculated
    in comparison to the `refractive_index_medium` property of the optical device.

    Parameters
    ----------
    radius : float
        Radius of the mie particle in meter.
    refractive_index : float
        Refractive index of the particle
    L : int or str
        The number of terms used to evaluate the mie theory. If `"auto"`,
        it determines the number of terms automatically.
    position : array_like[float, float (, float)]
        The position of the particle. Third index is optional,
        and represents the position in the direction normal to the
        camera plane.
    z : float
        The position in the direction normal to the
        camera plane. Used if `position` is of length 2.
    value : float
        A default value of the characteristic of the particle. Used by
        optics unless a more direct property is set: (eg. `refractive_index`
        for `Brightfield` and `intensity` for `Fluorescence`).
    """

    def __init__(
        self,
        radius=1e-6,
        refractive_index=1.45,
        offset_z=lambda radius: max(radius * 2, 5e-6),
        polarization_angle=0,
        aperature_angle="auto",
        L="auto",
        crop_empty=False,
        **kwargs
    ):
        super().__init__(
            radius=radius,
            refractive_index=refractive_index,
            L=L,
            offset_z=offset_z,
            polarization_angle=polarization_angle,
            aperature_angle=aperature_angle,
            crop_empty=crop_empty,
            **kwargs
        )

    def _process_properties(self, properties):

        properties = super()._process_properties(properties)

        if properties["L"] == "auto":
            v = 2 * np.pi * properties["radius"] / properties["wavelength"]
            properties["L"] = int(np.ceil(v + 4 * (v ** (1 / 3)) + 2))
        if properties["aperature_angle"] == "auto":
            properties["aperature_angle"] = np.sqrt(
                1 - properties["NA"] ** 2 / properties["refractive_index_medium"] ** 2
            )
        return properties

    def get(
        self,
        image,
        position,
        radius,
        refractive_index,
        upscaled_output_region,
        voxel_size,
        padding,
        wavelength,
        refractive_index_medium,
        L,
        offset_z,
        aperature_angle,
        polarization_angle,
        upscale=1,
        **kwargs
    ):

        xSize = (
            padding[2]
            + upscaled_output_region[2]
            - upscaled_output_region[0]
            + padding[0]
        )
        ySize = (
            padding[3]
            + upscaled_output_region[3]
            - upscaled_output_region[1]
            + padding[1]
        )
        arr = deeptrack.image.pad_image_to_fft(np.zeros((xSize, ySize)))

        x = (
            np.arange(-padding[0], arr.shape[0] - padding[0])
            - (position[1] + 0.5) * upscale
        )
        y = (
            np.arange(-padding[1], arr.shape[1] - padding[1])
            - (position[0] + 0.5) * upscale
        )
        X, Y = np.meshgrid(x * voxel_size[0], y * voxel_size[1])

        ct_max = np.cos(aperature_angle)
        field = _get_field(
            X,
            Y,
            offset_z,
            2 * np.pi / wavelength,
            refractive_index,
            radius,
            L,
            refractive_index_medium,
            ct_max=ct_max,
            polarization_angle=polarization_angle,
        )

        return np.expand_dims(field, axis=-1)


CACHE = {
    "maxL": -1,
    "minKR": None,
    "maxKR": 0,
    "resolution": 0.001,
    "L": [],
    "PI": {
        "value": [],
        "resolution": 0.001,
        "min": -1,
        "max": 1,
    },
}


def B(k, n, a, l, nm=1.33):
    ka = k * nm * a
    kna = k * n * a
    jka = jv(l, ka)
    djka = jv(l, ka, True)
    jkna = jv(l, kna)
    djkna = jv(l, kna, True)
    h = h1vp(l, ka, n=1)
    return (
        (2 * l + 1)
        * 1j ** l
        * (jka * djkna * n - jkna * djka * nm)
        / (jkna * h * nm - h * djkna * n)
    )


def ricbesj(l, x):
    return np.sqrt(np.pi * x / 2) * besselj(l + 0.5, x)


def dricbesj(l, x):
    return 0.5 * np.sqrt(np.pi / x / 2) * besselj(l + 0.5, x) + np.sqrt(
        np.pi * x / 2
    ) * dbesselj(l + 0.5, x)


def besselj(l, x):
    return jn(l, x)


def dbesselj(l, x):
    return 0.5 * (besselj(l - 1, x) - besselj(l + 1, x))


def ricbesh(l, x):
    return np.sqrt(np.pi * x / 2) * h1vp(l + 0.5, x, False)


def dricbesh(nu, z):
    xi = 0.5 * np.sqrt(np.pi / 2 / z) * h1vp(nu + 0.5, z, False) + np.sqrt(
        np.pi * z / 2
    ) * h1vp(nu + 0.5, z, True)
    return xi


def coeffs(k, n, a, L, nm=1.33):
    AA = np.zeros((L + 1,)) * 1j
    BB = np.zeros((L + 1,)) * 1j
    m = n / nm
    for l in range(1, L + 1):

        Sx = ricbesj(l, k * a)
        dSx = dricbesj(l, k * a)

        Smx = ricbesj(l, k * m * a)

        dSmx = dricbesj(l, k * m * a)
        xix = ricbesh(l, k * a)
        dxix = dricbesh(l, k * a)
        AA[l - 1] = (m * Smx * dSx - Sx * dSmx) / (m * Smx * dxix - xix * dSmx)
        BB[l - 1] = (Smx * dSx - m * Sx * dSmx) / (Smx * dxix - m * xix * dSmx)
    return AA, BB


def h1vp_cached(L, hin):
    hin_max = np.max(hin)
    if hin_max > CACHE["maxKR"]:
        calc_arr = np.arange(
            CACHE["maxKR"] + CACHE["resolution"],
            hin_max + CACHE["resolution"],
            CACHE["resolution"],
        )
        for l in range(CACHE["maxL"]):
            CACHE["L"][l] = np.concatenate((CACHE["L"][l], h1vp(l, calc_arr)))
        CACHE["maxKR"] = calc_arr[-1]

    if L > len(CACHE["L"]):
        calc_arr = np.arange(
            0, CACHE["maxKR"] + CACHE["resolution"], CACHE["resolution"]
        )
        for l in range(len(CACHE["L"]), L):
            CACHE["L"].append(h1vp(l, calc_arr))

        CACHE["maxL"] = L

    residx = ((hin // CACHE["resolution"])).astype(np.int)
    c = CACHE["L"]
    out = [c[l][residx] for l in range(L)]
    return out


def _get_scattering_matrix(A, B, k, L):
    theta = np.linspace(0, 2 * np.pi)


def _get_angular_dependence(ct, L):
    PI = np.zeros((L, *ct.shape))
    TAU = np.zeros((L, *ct.shape))

    PI[0, :] = 1
    PI[1, :] = 3 * ct
    TAU[0, :] = ct
    TAU[1, :] = 6 * ct * ct - 3

    for i in range(3, L + 1):

        PI[i - 1] = (2 * i - 1) / (i - 1) * ct * PI[i - 2] - i / (i - 1) * PI[i - 3]

        TAU[i - 1] = i * ct * PI[i - 1] - (i + 1) * PI[i - 2]
    return PI, TAU


def _get_field(X, Y, dz, k, n, a, L, nm=1.33, ct_max=1, polarization_angle=0):
    k = k * nm
    A, B = coeffs(k, n, a, L, nm)

    R2 = np.sqrt(X ** 2 + Y ** 2)
    R3 = np.sqrt(R2 ** 2 + (dz) ** 2)
    ct = dz / R3

    PI, TAU = _get_angular_dependence(ct, L)

    E = [(2 * l + 1) / (l * (l + 1)) for l in range(1, L + 1)]

    S1 = sum([E[l] * A[l] * TAU[l] + E[l] * B[l] * PI[l] for l in range(0, L)])
    S2 = sum([E[l] * B[l] * TAU[l] + E[l] * A[l] * PI[l] for l in range(0, L)])

    ANGLE = np.arctan2(Y, X) + polarization_angle
    COS = np.square(np.cos(ANGLE))
    SIN = 1 - COS

    field = (
        (ct > ct_max)
        * -1j
        / (k * R3)
        * np.exp(1j * k * (R3 - dz))
        * (S1 * COS + S2 * SIN)
    )

    return field



import sys

sys.path.append("..")  # Adds the module to path

import unittest

import deeptrack.aberrations as aberrations

from deeptrack.scatterers import PointParticle
from deeptrack.optics import Fluorescence
from deeptrack.image import Image


class TestAberrations(unittest.TestCase):

    particle = PointParticle(position=(32, 32), position_unit="pixel", intensity=1)

    def testGaussianApodization(self):
        aberrated_optics = Fluorescence(
            NA=0.3,
            resolution=1e-6,
            magnification=10,
            wavelength=530e-9,
            output_region=(0, 0, 64, 64),
            padding=(64, 64, 64, 64),
            aberration=aberrations.GaussianApodization(sigma=0.5),
        )
        aberrated_particle = aberrated_optics(self.particle)
        for z in (-100, 0, 100):
            im = aberrated_particle.resolve(z=z)
            self.assertIsInstance(im, Image)
            self.assertEqual(im.shape, (64, 64, 1))

    def testPiston(self):
        aberrated_optics = Fluorescence(
            NA=0.3,
            resolution=1e-6,
            magnification=10,
            wavelength=530e-9,
            output_region=(0, 0, 64, 64),
            padding=(64, 64, 64, 64),
            aberration=aberrations.Piston(coefficient=1),
        )
        aberrated_particle = aberrated_optics(self.particle)
        for z in (-100, 0, 100):
            im = aberrated_particle.resolve(z=z)
            self.assertIsInstance(im, Image)
            self.assertEqual(im.shape, (64, 64, 1))

    def testVerticalTilt(self):
        aberrated_optics = Fluorescence(
            NA=0.3,
            resolution=1e-6,
            magnification=10,
            wavelength=530e-9,
            output_region=(0, 0, 64, 64),
            padding=(64, 64, 64, 64),
            aberration=aberrations.VerticalTilt(coefficient=1),
        )
        aberrated_particle = aberrated_optics(self.particle)
        for z in (-100, 0, 100):
            im = aberrated_particle.resolve(z=z)
            self.assertIsInstance(im, Image)
            self.assertEqual(im.shape, (64, 64, 1))

    def testHorizontalTilt(self):
        aberrated_optics = Fluorescence(
            NA=0.3,
            resolution=1e-6,
            magnification=10,
            wavelength=530e-9,
            output_region=(0, 0, 64, 64),
            padding=(64, 64, 64, 64),
            aberration=aberrations.HorizontalTilt(coefficient=1),
        )
        aberrated_particle = aberrated_optics(self.particle)
        for z in (-100, 0, 100):
            im = aberrated_particle.resolve(z=z)
            self.assertIsInstance(im, Image)
            self.assertEqual(im.shape, (64, 64, 1))

    def testObliqueAstigmatism(self):
        aberrated_optics = Fluorescence(
            NA=0.3,
            resolution=1e-6,
            magnification=10,
            wavelength=530e-9,
            output_region=(0, 0, 64, 64),
            padding=(64, 64, 64, 64),
            aberration=aberrations.ObliqueAstigmatism(coefficient=1),
        )
        aberrated_particle = aberrated_optics(self.particle)
        for z in (-100, 0, 100):
            im = aberrated_particle.resolve(z=z)
            self.assertIsInstance(im, Image)
            self.assertEqual(im.shape, (64, 64, 1))

    def testDefocus(self):
        aberrated_optics = Fluorescence(
            NA=0.3,
            resolution=1e-6,
            magnification=10,
            wavelength=530e-9,
            output_region=(0, 0, 64, 64),
            padding=(64, 64, 64, 64),
            aberration=aberrations.Defocus(coefficient=1),
        )
        aberrated_particle = aberrated_optics(self.particle)
        for z in (-100, 0, 100):
            im = aberrated_particle.resolve(z=z)
            self.assertIsInstance(im, Image)
            self.assertEqual(im.shape, (64, 64, 1))

    def testAstigmatism(self):
        aberrated_optics = Fluorescence(
            NA=0.3,
            resolution=1e-6,
            magnification=10,
            wavelength=530e-9,
            output_region=(0, 0, 64, 64),
            padding=(64, 64, 64, 64),
            aberration=aberrations.Astigmatism(coefficient=1),
        )
        aberrated_particle = aberrated_optics(self.particle)
        for z in (-100, 0, 100):
            im = aberrated_particle.resolve(z=z)
            self.assertIsInstance(im, Image)
            self.assertEqual(im.shape, (64, 64, 1))

    def testObliqueTrefoil(self):
        aberrated_optics = Fluorescence(
            NA=0.3,
            resolution=1e-6,
            magnification=10,
            wavelength=530e-9,
            output_region=(0, 0, 64, 64),
            padding=(64, 64, 64, 64),
            aberration=aberrations.ObliqueTrefoil(coefficient=1),
        )
        aberrated_particle = aberrated_optics(self.particle)
        for z in (-100, 0, 100):
            im = aberrated_particle.resolve(z=z)
            self.assertIsInstance(im, Image)
            self.assertEqual(im.shape, (64, 64, 1))

    def testVerticalComa(self):
        aberrated_optics = Fluorescence(
            NA=0.3,
            resolution=1e-6,
            magnification=10,
            wavelength=530e-9,
            output_region=(0, 0, 64, 64),
            padding=(64, 64, 64, 64),
            aberration=aberrations.VerticalComa(coefficient=1),
        )
        aberrated_particle = aberrated_optics(self.particle)
        for z in (-100, 0, 100):
            im = aberrated_particle.resolve(z=z)
            self.assertIsInstance(im, Image)
            self.assertEqual(im.shape, (64, 64, 1))

    def testHorizontalComa(self):
        aberrated_optics = Fluorescence(
            NA=0.3,
            resolution=1e-6,
            magnification=10,
            wavelength=530e-9,
            output_region=(0, 0, 64, 64),
            padding=(64, 64, 64, 64),
            aberration=aberrations.HorizontalComa(coefficient=1),
        )
        aberrated_particle = aberrated_optics(self.particle)
        for z in (-100, 0, 100):
            im = aberrated_particle.resolve(z=z)
            self.assertIsInstance(im, Image)
            self.assertEqual(im.shape, (64, 64, 1))

    def testTrefoil(self):
        aberrated_optics = Fluorescence(
            NA=0.3,
            resolution=1e-6,
            magnification=10,
            wavelength=530e-9,
            output_region=(0, 0, 64, 64),
            padding=(64, 64, 64, 64),
            aberration=aberrations.Trefoil(coefficient=1),
        )
        aberrated_particle = aberrated_optics(self.particle)
        for z in (-100, 0, 100):
            im = aberrated_particle.resolve(z=z)
            self.assertIsInstance(im, Image)
            self.assertEqual(im.shape, (64, 64, 1))

    def testSphericalAberration(self):
        aberrated_optics = Fluorescence(
            NA=0.3,
            resolution=1e-6,
            magnification=10,
            wavelength=530e-9,
            output_region=(0, 0, 64, 64),
            padding=(64, 64, 64, 64),
            aberration=aberrations.SphericalAberration(coefficient=1),
        )
        aberrated_particle = aberrated_optics(self.particle)
        for z in (-100, 0, 100):
            im = aberrated_particle.resolve(z=z)
            self.assertIsInstance(im, Image)
            self.assertEqual(im.shape, (64, 64, 1))


if __name__ == "__main__":
    unittest.main()


"""Tools to manage the properties of a feature

Classes
-------
Property
    The class `Property`, which represents the values of a property of a feature.
    A Property be:
    * A constant (initialization with, e.g., a number, a tuple)
    * A sequence of variables (initialization with, e.g., an iterator)
    * A random variable (initialization with, e.g., a function)
SequentialProperty
    The class `SequentialProperty`, which extends `Property` to sample one value
    for each step in a sequence.
PropertyDict
    The class `PropertyDict`, which is a dictionary with each element a Property.
    The class provides utility functions to update, sample, clear and retrieve
    properties.
"""

import numpy as np
from deeptrack.utils import isiterable, hasmethod, get_kwarg_names, kwarg_has_default
import deeptrack
import copy


class Property:
    """Represents a property of a feature

    The class Property` wraps an input, which is treated
    internally as a sampling rule. This sampling rule is used
    to update the value of the property.
    The sampling rule can be, for example:
    * A constant (initialization with, e.g., a number, a tuple)
    * A sequence of variables (initialization with, e.g., a generator)
    * A discrete random variable (initialization with, e.g., a list, a dictionary)
    * A continuous random variable (initialization with, e.g., a function)

    Parameters
    ----------
    sampling_rule : any
        Defines the sampling rule to update the value of the feature property.
        See method `sample()` for how different sampling rules are sampled.

    Attributes
    ----------
    sampling_rule : any
        The sampling rule to update the value of the feature property.
    current_value : any
        The current value obtained from the last call to the sampling rule.

    """

    def __init__(self, sampling_rule: any):
        self.sampling_rule = sampling_rule
        self.parent = None

    @property
    def current_value(self):
        """Current value of the property of the feature

        `current_value` is the result of the latest `update()` call.
        Note that any randomization only occurs when the method `update()` is called
        and, therefore, the current value does not change between calls.

        The method getter calls the method `update()` if `current_value`
        has not been set yet.

        """

        self._current_value

    @current_value.setter
    def current_value(self, updated_current_value):
        self._current_value = updated_current_value
        if id(self) not in deeptrack.UPDATE_MEMO["memoization"]:
            # Some values work, some don't. self, updated_current_value and self._current_value work
            # Best guess is an error in the gc reference counter causing it to dereference
            # But then again, I don't think it should be the same reference anyway
            deeptrack.UPDATE_MEMO["memoization"][id(self)] = updated_current_value

    @current_value.getter
    def current_value(self):

        if not hasattr(self, "_current_value"):
            self.update()

        return self._current_value

    def update(self, **kwargs) -> "Property":
        """Updates the current value

        The method `update()` sets the property `current_value`
        as the output of the method `sample()`. Will only update
        once per resolve.

        Any object that implements the method `update()` will have it called.

        Returns
        -------
        Property
            Returns itself.

        """

        # If currently updated through a call to feature and deeptrack.UPDATE_MEMO
        my_id = id(self)

        # if my_id in deeptrack.UPDATE_MEMO["memoization"] and not hasattr(self, '_current_value'):
        #     a = 1+1

        if (
            deeptrack.UPDATE_LOCK.locked()
            and my_id in deeptrack.UPDATE_MEMO["memoization"]
        ):
            return self

        if self.parent:
            kwargs.update(self.parent)

        kwargs.update(deeptrack.UPDATE_MEMO["user_arguments"])
        self.current_value = self.sample(self.sampling_rule, **kwargs)

        return self

    def sample(self, sampling_rule, **kwargs):
        """Samples the sampling rule

        Returns a sampled instance of the `sampling_rule` field.
        The logic behind the sampling depends on the type of
        `sampling_rule`. These are checked in the following order of
        priority:

        1. Any object with a callable `sample()` method has this
            method called and returned.
        2. If the rule is a ``dict``, sample each value and combine the
            result into a new ``dict`` using the original keys.
        3. If the rule is a ``list``, sample each element of the list and
            combine the result into a ne ``list``.
        4. If the rule is an ``iterable``, return the next output.
        5. If the rule is callable, call it with its accepted arguments.
            Example arguments can be the value of some other property.
        6. If none of the above apply, return the rule itself.

        Parameters
        ----------
        sampling_rule : any
            The rule to sample values from.
        **kwargs
            Arguments that will be passed on to functions that accepts them.

        Returns
        -------
        any
            A sampled instance of the `sampling_rule`.

        """

        if isinstance(sampling_rule, deeptrack.Feature):
            # Don't pass my properties to other feature (avoid name clash)
            sampling_rule._update()
            return sampling_rule

        if isinstance(sampling_rule, Property):
            sampling_rule.parent.update_item(sampling_rule)
            return sampling_rule.current_value

        elif isinstance(sampling_rule, dict):
            # If the ruleset is a dict, return a new dict with each
            # element being sampled from the original dict.
            out = {}
            for key, val in sampling_rule.items():
                out[key] = self.sample(val, **kwargs)
            return out

        elif isinstance(sampling_rule, list):
            return [self.sample(item, **kwargs) for item in sampling_rule]

        elif isinstance(sampling_rule, (tuple, np.ndarray)):
            # tuple and ndarrays are elementary
            return sampling_rule

        elif isiterable(sampling_rule):
            # If it's iterable, return the next value
            try:
                return next(sampling_rule)
            except StopIteration:
                return self.current_value

        elif callable(sampling_rule):
            # If it's a function, extract the arguments it accepts.
            function_input = {}

            # Get the kwarg arguments the function accepts
            knames = get_kwarg_names(sampling_rule)
            for i in range(len(knames)):
                key = knames[i]
                # If that name is among passed kwarg arguments
                if key in kwargs:
                    if isinstance(kwargs[key], Property):
                        # If it is a property, update it and pass the current value
                        if not kwargs[key] is self:
                            if kwargs[key].parent:
                                kwargs[key].parent.update_item(kwargs[key])
                            else:
                                kwargs[key].update(**kwargs)

                        if (
                            isinstance(kwargs[key], SequentialProperty)
                            and "sequence_step" in kwargs
                        ):
                            function_input[key] = kwargs[key].current_value[
                                kwargs["sequence_step"]
                            ]
                        else:
                            function_input[key] = kwargs[key].current_value

                    else:
                        function_input[key] = kwargs[key]

                elif not kwarg_has_default(sampling_rule, key):
                    function_input[key] = None

            new_value = sampling_rule(**function_input)
            while isinstance(new_value, Property):
                new_value = self.sample(new_value, **kwargs)
            return new_value
        else:
            # Else, assume it's elementary.
            return sampling_rule

    def __deepcopy__(self, memo):
        is_in = id(self) in deeptrack.UPDATE_MEMO["memoization"]
        if is_in:
            return self
        else:
            cls = self.__class__  # Extract the class of the object
            result = cls.__new__(
                cls
            )  # Create a new instance of the object based on extracted class
            memo[id(self)] = result
            for k, v in self.__dict__.items():
                setattr(
                    result, k, copy.deepcopy(v, memo)
                )  # Copy over attributes by copying directly or in case of complex objects like lists for exaample calling the `__deepcopy()__` method defined by them. Thus recursively copying the whole tree of objects.
            return result


class SequentialProperty(Property):
    """Property that has multiple sequential values

    Extends standard `Property` to resolve one value for each step
    in a sequence of images. This is often used when creating movies.

    Parameters
    ----------
    initializer : any
        Sampling rule for the first step of the sequence.
    sampling_rule : any
        Sampling rule for each step after the first.

    Attributes
    ----------
    initializer : any
        Sampling rule for the first step of the sequence.
    sampling_rule : any
        Sampling rule for each step after the first.
    has_updated_since_last_resolve : bool
        Whether the property has been updated since the last resolve.

    """

    def __init__(self, sampling_rule, initializer=None):
        super().__init__(sampling_rule)
        if initializer is None:
            self.initializer = sampling_rule
        else:
            self.initializer = initializer

        self.sampling_rule = sampling_rule

        # Deprecated
        self.has_updated_since_last_resolve = False

    def update(self, sequence_length=0, **kwargs):
        """Updates current_value

        For each step in the sequence, sample `self.sampling_rule`.
        `self.initializer` is used for the first step. These rules
        should output one value per step. Sampling rules
        that are functions can optionally accept the following keyword
        arguments:

        * sequence_step : the current position in the sequence.
        * sequence_length : the length of the sequence.
        * previous_value : the value of the property at the previous
          step in the sequence.
        * previous_values : the value of the property at all previous
          steps in the sequence.

        Parameters
        ----------
        sequence_length : int, optional
            length of the sequence

        Returns
        -------
        self
            returns self
        """
        my_id = id(self)
        if (
            deeptrack.UPDATE_LOCK.locked()
            and my_id in deeptrack.UPDATE_MEMO["memoization"]
        ):
            return self

        kwargs.update(deeptrack.UPDATE_MEMO["user_arguments"])

        new_current_value = []

        for step in range(sequence_length):
            # Use initializer for first time step
            ruleset = self.sampling_rule

            # Elements inserted here can be passed to property functions
            kwargs.update(
                sequence_step=step,
                sequence_length=sequence_length,
                previous_values=new_current_value,
            )
            if step == 0:
                kwargs.update(previous_value=self.sample(self.initializer, **kwargs))
            else:
                kwargs.update(previous_value=new_current_value[-1])

            next_value = self.sample(ruleset, **kwargs)

            new_current_value.append(next_value)

        self.current_value = new_current_value
        deeptrack.UPDATE_MEMO["memoization"][my_id] = new_current_value
        return self


import collections


class PropertyDict(collections.OrderedDict):
    """Dictionary with Property elements

    A dictionary of properties. It provides utility functions to update,
    sample, reset and retrieve properties.

    Parameters
    ----------
    *args, **kwargs
        Arguments used to initialize a dict

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for k, v in self.items():
            if hasattr(v, "parent") and not v.parent:
                v.parent = self

    def current_value_dict(self, **kwargs) -> dict:
        """Retrieves the current value of all properties as a dictionary

        Returns
        -------
        dict
            A dictionary with the current value of all properties

        """
        current_value_dict = {}
        for key, property in self.items():

            property_value = property.current_value

            # If the property is sequential, retrieve the value
            # of the current timestep
            if isinstance(property, SequentialProperty):
                sequence_step = kwargs.get("sequence_step", None)
                if not sequence_step is None:
                    property_value = property_value[sequence_step]

            current_value_dict[key] = property_value

        return current_value_dict

    def update(self, **kwargs) -> "PropertyDict":
        """Updates all properties

        Calls the method `update()` on each property in the dictionary.

        Returns
        -------
        Properties
            Returns itself

        """
        property_arguments = collections.OrderedDict(self)
        property_arguments.update(kwargs)
        property_arguments.update(deeptrack.UPDATE_MEMO["user_arguments"])
        for key, prop in self.items():
            if isinstance(property_arguments[key], Property):
                prop.update(**property_arguments)
            elif id(prop) not in deeptrack.UPDATE_MEMO["memoization"]:
                prop.current_value = property_arguments[key]

        return self

    def update_item(self, item: Property, **kwargs):
        """Updates a single property.

        Parameters
        ----------
        item: Property
            The item to update.

        Returns
        -------
        self
        """
        property_arguments = collections.OrderedDict(self)
        property_arguments.update(kwargs)
        for key, prop in self.items():
            if prop == item:
                prop.update(**property_arguments)
                break

        return self

    def sample(self, **kwargs) -> dict:
        """Samples all properties

        Returns
        -------
        dict
            A dictionary with each key-value pair the result of a
            `sample()` call on the property with the same key.

        """

        sample_dict = {}
        for key, property in self.items():
            sample_dict[key] = property.sample(**kwargs)

        return sample_dict


import sys

sys.path.append("..")  # Adds the module to path

import unittest

import deeptrack.optics as optics

from deeptrack.scatterers import PointParticle
from deeptrack.image import Image


class TestOptics(unittest.TestCase):
    def test_Fluorescence(self):
        microscope = optics.Fluorescence(
            NA=0.7,
            wavelength=660e-9,
            resolution=1e-6,
            magnification=10,
            refractive_index_medium=1.33,
            upscale=2,
            padding=(10, 10, 10, 10),
            output_region=(0, 0, 64, 64),
            aberration=None,
        )
        scatterer = PointParticle(
            intensity=100,  # Squared magnitude of the field.
            position_unit="pixel",  # Units of position (default meter)
            position=(32, 32),  # Position of the particle
        )
        imaged_scatterer = microscope(scatterer)
        output_image = imaged_scatterer.resolve()
        self.assertIsInstance(output_image, Image)
        self.assertEqual(output_image.shape, (64, 64, 1))

    def test_Brightfield(self):
        microscope = optics.Brightfield(
            NA=0.7,
            wavelength=660e-9,
            resolution=1e-6,
            magnification=10,
            refractive_index_medium=1.33,
            upscale=2,
            output_region=(0, 0, 64, 64),
            padding=(10, 10, 10, 10),
            aberration=None,
        )
        scatterer = PointParticle(
            refractive_index=1.45 + 0.1j,
            position_unit="pixel",
            position=(32, 32),
        )
        imaged_scatterer = microscope(scatterer)
        output_image = imaged_scatterer.resolve()
        self.assertIsInstance(output_image, Image)
        self.assertEqual(output_image.shape, (64, 64, 1))

    def test_IlluminationGradient(self):
        illumination_gradient = optics.IlluminationGradient(gradient=(5e-5, 5e-5))
        microscope = optics.Brightfield(
            NA=0.7,
            wavelength=660e-9,
            resolution=1e-6,
            magnification=10,
            refractive_index_medium=1.33,
            upscale=2,
            output_region=(0, 0, 64, 64),
            padding=(10, 10, 10, 10),
            aberration=None,
            illumination=illumination_gradient,
        )
        scatterer = PointParticle(
            refractive_index=1.45 + 0.1j,
            position_unit="pixel",
            position=(32, 32),
        )
        imaged_scatterer = microscope(scatterer)
        output_image = imaged_scatterer.resolve()
        self.assertIsInstance(output_image, Image)
        self.assertEqual(output_image.shape, (64, 64, 1))


if __name__ == "__main__":
    unittest.main()


import sys

sys.path.append("..")  # Adds the module to path

import unittest

import deeptrack.image as image

from deeptrack.features import Feature
import numpy as np


class TestImage(unittest.TestCase):
    class Particle(Feature):
        def get(self, image, position=None, **kwargs):
            # Code for simulating a particle not included
            return image

    def test_Image(self):
        particle = self.Particle(position=(128, 128))
        input_image = image.Image(np.zeros((256, 256)))
        output_image = particle.resolve(input_image)
        self.assertIsInstance(output_image, image.Image)

    def test_Image_properties(self):
        particle = self.Particle(position=(128, 128))
        input_image = image.Image(np.zeros((256, 256)))
        output_image = particle.resolve(input_image)
        properties = output_image.properties
        self.assertIsInstance(properties, list)
        self.assertIsInstance(properties[0], dict)
        self.assertEqual(properties[0]["position"], (128, 128))
        self.assertEqual(properties[0]["name"], "Particle")

    def test_Image_get_property(self):
        particle = self.Particle(position=(128, 128))
        input_image = image.Image(np.zeros((256, 256)))
        output_image = particle.resolve(input_image)

        property_position = output_image.get_property("position")
        self.assertEqual(property_position, (128, 128))

        property_name = output_image.get_property("name")
        self.assertEqual(property_name, "Particle")

        property_hash_key = output_image.get_property("hash_key")
        self.assertIsInstance(property_hash_key, list)

    def test_Image_append(self):
        particle = self.Particle(position=(128, 128))
        input_image = image.Image(np.zeros((256, 256)))
        output_image = particle.resolve(input_image)
        properties = output_image.properties

        property_dict = {"key1": 1, "key2": 2}
        output_image.append(property_dict)
        self.assertEqual(properties[0]["position"], (128, 128))
        self.assertEqual(properties[0]["name"], "Particle")
        self.assertEqual(properties[1]["key1"], 1)
        self.assertEqual(output_image.get_property("key1"), 1)
        self.assertEqual(properties[1]["key2"], 2)
        self.assertEqual(output_image.get_property("key2"), 2)

        property_dict2 = {"key1": 11, "key2": 22}
        output_image.append(property_dict2)
        self.assertEqual(output_image.get_property("key1"), 1)
        self.assertEqual(output_image.get_property("key1", get_one=False), [1, 11])

    def test_Image_merge_properties_from(self):
        particle = self.Particle(position=(128, 128))
        input_image = image.Image(np.zeros((256, 256)))
        output_image1 = particle.resolve(input_image)
        output_image2 = particle.resolve(input_image)
        output_image1.merge_properties_from(output_image2)
        self.assertEqual(len(output_image1.properties), 1)

        particle.update()
        output_image3 = particle.resolve(input_image)
        output_image1.merge_properties_from(output_image3)
        self.assertEqual(len(output_image1.properties), 2)

    def test_pad_image_to_fft(self):
        input_image = image.Image(np.zeros((7, 25)))
        padded_image = image.pad_image_to_fft(input_image)
        self.assertEqual(padded_image.shape, (8, 27))

        input_image = image.Image(np.zeros((30, 27)))
        padded_image = image.pad_image_to_fft(input_image)
        self.assertEqual(padded_image.shape, (32, 27))

        input_image = image.Image(np.zeros((300, 400)))
        padded_image = image.pad_image_to_fft(input_image)
        self.assertEqual(padded_image.shape, (324, 432))


if __name__ == "__main__":
    unittest.main()


import sys

sys.path.append("..")  # Adds the module to path

import unittest

import deeptrack.properties as properties
import deeptrack as dt

import numpy as np


class TestProperties(unittest.TestCase):
    def test_Property_constant(self):
        P = properties.Property(1)
        self.assertEqual(P.current_value, 1)
        P.update()
        self.assertEqual(P.current_value, 1)

    def test_Property_iter(self):
        P = properties.Property(iter([1, 2, 3, 4, 5]))
        self.assertEqual(P.current_value, 1)
        for i in range(1, 5):
            self.assertEqual(P.current_value, i)
            P.update()

    def test_Property_random(self):
        P = properties.Property(lambda: np.random.rand())
        for _ in range(100):
            P.update()
            self.assertTrue(P.current_value >= 0 and P.current_value <= 1)

    def test_SequentialProperty_Constant(self):
        sampling_rule = 1
        P = properties.SequentialProperty(sampling_rule, initializer=0)
        P.update(sequence_length=10)
        self.assertEqual(P.current_value, [1, 1, 1, 1, 1, 1, 1, 1, 1, 1])

    def test_SequentialProperty_timedependent(self):
        def steady_increase(previous_value):
            return previous_value + 1

        sampling_rule = steady_increase
        P = properties.SequentialProperty(sampling_rule, initializer=2)
        P.update(sequence_length=5)
        self.assertEqual(P.current_value, [3, 4, 5, 6, 7])

    def test_SequentialProperty_dependence_on_SequentialProperty(self):
        def steady_increase(previous_value):
            return previous_value + 1

        P1 = properties.SequentialProperty(steady_increase, initializer=1)

        def geometric_increase(previous_value, step_length):
            return previous_value + step_length

        P2 = properties.SequentialProperty(geometric_increase, initializer=1)

        P2.update(sequence_length=5, step_length=P1)
        self.assertEqual(P2.current_value, [3, 6, 10, 15, 21])

    def test_UpdateFromParent(self):
        optics = dt.Fluorescence()
        sample = dt.DummyFeature(voxel_size=optics.voxel_size)
        together = optics(sample)
        together.update()
        self.assertEqual(
            tuple(optics.voxel_size.current_value),
            tuple(sample.voxel_size.current_value),
        )

    def test_PropertyDict(self):
        property_dict = properties.PropertyDict(
            P1=properties.Property(1),
            P2=properties.Property(iter([1, 2, 3, 4, 5])),
            P3=properties.Property(lambda: np.random.rand()),
        )
        current_value_dict = property_dict.current_value_dict()
        self.assertEqual(current_value_dict["P1"], 1)
        self.assertEqual(current_value_dict["P2"], 1)
        self.assertTrue(current_value_dict["P3"] >= 0 and current_value_dict["P3"] <= 1)
        for i in range(1, 100):
            current_value_dict = property_dict.current_value_dict(is_resolving=True)
            self.assertEqual(current_value_dict["P1"], 1)
            self.assertEqual(current_value_dict["P2"], np.min((i, 5)))
            self.assertTrue(
                current_value_dict["P3"] >= 0 and current_value_dict["P3"] <= 1
            )
            property_dict.update()


if __name__ == "__main__":
    unittest.main()


""" Loss functions specialized for images and tracking tasks.

Functions
---------
flatten 
    Flattends the inputs before calling the loss function.
sigmoid
    Adds a signmoid transformation to the prediction before calling the loss function.
weighted_crossentropy
    Binary crossentropy with weighted classes.
nd_mean_squared_error
    Mean square error with flattened inputs.
nd_mean_squared_logarithmic_error
    Mean square log error with flattened inputs.
nd_poisson
    Poisson error loss flattened inputs.
nd_squared_hinge
    Squared hinge error with flattened inputs.
nd_binary_crossentropy
    Binary crossentropy error with flattened inputs.
nd_kullback_leibler_divergence
    Kullback-Leibler divergence error with flattened inputs.
nd_mean_absolute_error
    Mean absolute error with flattened inputs.
nd_mean_absolute_percentage_error
    Mean absolute percentage error with flattened inputs.
"""

import tensorflow.keras as keras

losses = keras.losses
K = keras.backend


# Extends the following loss functions
_COMPATIBLE_LOSS_FUNCTIONS = [
    losses.mse,
    losses.msle,
    losses.poisson,
    losses.squared_hinge,
    losses.binary_crossentropy,
    losses.kld,
    losses.mae,
    losses.mape,
]


# LOSS WRAPPERS
def flatten(func):
    """Flattens the inputs before calling the loss function.

    Parameters
    ----------
    func : loss function
        The loss function to wrap.

    Returns
    -------
    function
        The new loss function.
    """

    def wrapper(T, P):
        T = K.flatten(T)
        P = K.flatten(P)
        return func(T, P)

    wrapper.__name__ = "nd_" + func.__name__
    return wrapper


def sigmoid(func):
    """Adds a signmoid transformation to the prediction before calling the loss function.

    Parameters
    ----------
    func : loss function
        The loss function to wrap.

    Returns
    -------
    function
        The new loss function.
    """
    # Takes the Sigmoid function of the prediction
    def wrapper(T, P):
        P = K.clip(P, -50, 50)
        P = 1 / (1 + K.exp(-P))
        return func(T, P)

    wrapper.__name__ = "sigmoid_" + func.__name__
    return wrapper


def weighted_crossentropy(weight=(1, 1), eps=1e-4):
    """Binary crossentropy with weighted classes.

    Parameters
    ----------
    weight : Tuple[float, float]
        Tuple of two numbers, indicating the weighting of the two classes -- 1 and 0.

    Returns
    -------
    function
        Weighted binary crossentropy loss function
    """

    def unet_crossentropy(T, P):
        return -K.mean(
            weight[0] * T * K.log(P + eps) + weight[1] * (1 - T) * K.log(1 - P + eps)
        ) / (weight[0] + weight[1])

    return unet_crossentropy


# Wrap standard keras loss function with flatten.
for keras_loss_function in _COMPATIBLE_LOSS_FUNCTIONS:
    deeptrack_loss_function = flatten(keras_loss_function)
    globals()[deeptrack_loss_function.__name__] = deeptrack_loss_function


"""Features and tools for resolving sequences of images.

Classes
-------
Sequence
    Resolves a feature as a sequence.

Functions
---------
Sequential
    Converts a feature to be resolved as a sequence.
"""

from deeptrack.features import Feature
from deeptrack.properties import SequentialProperty


class Sequence(Feature):
    """Resolves a feature as a sequence.

    The input feature is resolved `sequence_length` times, with the kwarg
    arguments `sequene_length` and `sequence_step` passed to all properties
    of the feature set.

    Parameters
    ----------
    feature : Feature
        The feature to resolve as a sequence.
    sequence_length : int
        The number of times to resolve the feature.

    Attributes
    ----------
    feature : Feature
        The feature to resolve as a sequence.
    """

    __distributed__ = False

    def __init__(self, feature: Feature, sequence_length: int = 1, **kwargs):
        self.feature = feature
        super().__init__(sequence_length=sequence_length, **kwargs)

        # Require update
        self.update()

    def get(self, input_list, sequence_length=None, **kwargs):
        return (input_list or []) + [
            self.feature.resolve(
                sequence_length=sequence_length, sequence_step=sequence_step, **kwargs
            )
            for sequence_step in range(sequence_length)
        ]

    def update(self, **kwargs):
        super().update(**kwargs)

        sequence_length = self.properties["sequence_length"].current_value
        if "sequence_length" not in kwargs:
            kwargs["sequence_length"] = sequence_length

        self.feature.update(**kwargs)

        return self


def Sequential(feature: Feature, **kwargs):
    """Converts a feature to be resolved as a sequence.

    Should be called on individual features, not combinations of features. All keyword
    arguments will be trated as sequential properties and will be passed to the parent feature.

    If a property from the keyword argument already exists on the feature, the existing property
    will be used to initilize the passed property (that is, it will be used for the first timestep).

    Parameters
    ----------
    feature : Feature
        Feature to make sequential.
    kwargs
        Keyword arguments to pass on as sequential properties of `feature`.

    """

    for property_name, sampling_rule in kwargs.items():

        if property_name in feature.properties:
            initializer = feature.properties[property_name].sampling_rule
        else:
            initializer = sampling_rule

        feature.properties[property_name] = SequentialProperty(
            sampling_rule, initializer=initializer
        )
        feature.properties[property_name].parent = feature.properties

    return feature



""" Image class and relative functions

Defines the Image class and functions that operate on it.

CLASSES
---------
Image
    Subclass of numpy `ndarray`. Has the additional attribute properties
    which contains the properties used to generate it as a `list` of
    `dicts`.

Functions
---------
pad_image_to_fft(image: Image, axes = (0, 1))
    Pads the image with zeros to optimize the speed of Fast Fourier
    Transforms.
"""

import numpy as np


class Image(np.ndarray):
    """Subclass of numpy ndarray

    The class Image is used by features to resolve images and store
    the current values of the properties of each feature in the feature
    series. These properties are stored in the field `properties`
    as a list of dictionaries, in the same order as that in which
    the features have been evaluated.

    The field `properties` is used to store and extract information
    about how an image has been generated.

    Parameters
    ----------
    input_array : array_like
        An array_like object that is used to instantiate the ndarray.
    properties : list of dicts, optional
        Optional parameter to set as the initial value for the field properties.

    Attributes
    ----------
    properties : list
        List of dictionaries of the current value of all properties of
        the features used to resolve the image.

    """

    # Used by numpy to determine output type of u_funcs.
    # This ensures that the output will always be an Image
    __array_priority__ = 999

    def __new__(cls, input_array, properties=None):
        # Converts input to ndarray, and then to an Image
        # In particular, it creates the properties

        image = np.array(input_array).view(cls)
        if properties is None:
            # If input_array has properties attribute, retrieve a copy of it
            properties = getattr(input_array, "properties", [])[:]
        image.properties = properties

        return image

    def append(self, property_dict: dict):
        """Appends a dictionary to the properties list.

        Parameters
        ----------
        property_dict : dict
            A dictionary to append to the property list. Most commonly
            the current values of the properties of a feature.

        Returns
        -------
        Image
            Returns itself.
        """

        self.properties.append(property_dict)
        return self

    def get_property(
        self, key: str, get_one: bool = True, default: any = None
    ) -> list or any:
        """Retrieve a property.

        If the feature has the property defined by `key`, return
        its current_value. Otherwise, return `default`.
        If `get_one` is True, the first instance is returned;
        otherwise, all instances are returned as a list.

        Parameters
        ----------
        key
            The name of the property.
        get_one: optional
            Whether to return all instances of that property or just the first.
        default : optional
            What is returned if the property is not found.

        Returns
        -------
        any
            The value of the property if found, else `default`.

        """

        if get_one:
            for prop in self.properties:
                if key in prop:
                    return prop[key]
            return default
        else:
            return [prop[key] for prop in self.properties if key in prop] or default

    def merge_properties_from(self, other: "Image") -> "Image":
        """Merge properties with those from another Image.

        Appends properties from another images such that no property is duplicated.
        Uniqueness of a dictionary of properties is determined from the
        property `hash_key`.

        Most functions involving two images should automatically output an image with
        merged properties. However, since each property is guaranteed to be unique,
        it is safe to manually call this function if there is any uncertainty.

        Parameters
        ----------
        other
            The Image to retrieve properties from.

        """

        for new_prop in other.properties:

            # If no hash_key, add it
            if "hash_key" not in new_prop:
                self.append(new_prop)
                continue

            should_append = True
            # Else, see if hash is unique
            for my_prop in self.properties:

                if (
                    "hash_key" in my_prop
                    and my_prop["hash_key"] == new_prop["hash_key"]
                ):

                    # Key is not unique, don't add
                    should_append = False
                    break

            if should_append:
                self.append(new_prop)

        return self

    def __array_wrap__(self, image_after_function, context=None):
        # Called at end when a function is called on an image
        # It might be that the information about properties is lost,
        # this method restores it.
        # This method also correctly concatenate the the properties of two images.

        if image_after_function is self:  # for in-place operations
            image_with_restored_properties = image_after_function
        else:
            image_with_restored_properties = Image(image_after_function)

        if context is not None:
            # context is information about operation

            func, args, _ = context
            input_args = args[: func.nin]

            for arg in input_args:

                if not arg is self and isinstance(arg, Image):
                    self.merge_properties_from(arg)

        return image_with_restored_properties

    def __array_finalize__(self, image):
        # Called when an image is created
        # It might be that the information about properties is lost,
        # this method restores it.

        if image is None:
            return

        # Ensure self has properties defined
        self.properties = getattr(self, "properties", [])

        # Merge from image if image is Image
        if isinstance(image, Image):
            self.merge_properties_from(image)


FASTEST_SIZES = [0]
for n in range(1, 10):
    FASTEST_SIZES += [2 ** a * 3 ** (n - a - 1) for a in range(n)]
FASTEST_SIZES = np.sort(FASTEST_SIZES)


def pad_image_to_fft(image: Image, axes=(0, 1)) -> Image:
    """Pads image to speed up fast fourier transforms.
    Pads image to speed up fast fourier transforms by adding 0s to the
    end of the image.

    Parameters
    ----------
    image
        The image to pad
    axes : iterable of int, optional
        The axes along which to pad.
    """

    def _closest(dim):
        # Returns the smallest value frin FASTEST_SIZES
        # larger than dim
        for size in FASTEST_SIZES:
            if size >= dim:
                return size

    new_shape = np.array(image.shape)
    for axis in axes:
        new_shape[axis] = _closest(new_shape[axis])

    increase = np.array(new_shape) - image.shape
    pad_width = [(0, inc) for inc in increase]

    return np.pad(image, pad_width, mode="constant")



""" Features that augment images

Augmentations are features that can resolve more than one image without
calling `.resolve()` on the parent feature. Specifically, they create
`updates_per_reload` images, while calling their parent feature
`load_size` times.

Classes
-------
Augmentation
    Base abstract augmentation class.
PreLoad
    Simple storage with no augmentation.
FlipLR
    Flips images left-right.
FlipUD
    Flips images up-down.
FlipDiagonal
    Flips images diagonally.
"""

from deeptrack.features import Feature
from deeptrack.image import Image
import numpy as np
from typing import Callable
from scipy.ndimage.interpolation import map_coordinates
from scipy.ndimage.filters import gaussian_filter
import warnings


class Augmentation(Feature):
    """Base abstract augmentation class.

    Augmentations are features that can resolve more than one image without
    calling `.resolve()` on the parent feature. Specifically, they create
    `updates_per_reload` images, while calling their parent feature
    `load_size` times. They achieve this by resolving `load_size` results
    from the parent feature at once, and randomly drawing one of these
    results as input to the method `.get()`. A new input is chosen
    every time `.update()` is called. Once `.update()` has been called
    `updated_per_reload` times, a new batch of `load_size` results are
    resolved from the parent feature.

    The method `.get()` of implementations of this class may accept the
    property `number_of_updates` as an argument. This number represents
    the number of times the `.update()` method has been called since the
    last time the parent feature was resolved.

    Parameters
    ----------
    feature : Feature, optional, deprecated
        DEPRECATED. The parent feature. If not passed, it acts like any other feature.
    load_size : int
        Number of results to resolve from the parent feature.
    updates_per_reload : int
        Number of times `.update()` is called before resolving new results
        from the parent feature.
    update_properties : Callable or None
        Function called on the output of the method `.get()`. Overrides
        the default behaviour, allowing full control over how to update
        the properties of the output to account for the augmentation.
    """

    __distributed__ = False

    def __init__(
        self,
        feature: Feature = None,
        load_size: int = 1,
        updates_per_reload: int = 1,
        update_properties: Callable or None = None,
        **kwargs
    ):

        if load_size is not 1:
            warnings.warn(
                "Using an augmentation with a load size other than one is no longer supported",
                DeprecationWarning,
            )

        self.feature = feature

        def get_number_of_updates(updates_per_reload=1):
            # Updates the number of updates. The very first update is not counted.
            if not hasattr(self.properties["number_of_updates"], "_current_value"):
                return 0
            return (
                self.properties["number_of_updates"].current_value + 1
            ) % updates_per_reload

        def tally():
            idx = 0
            while True:
                yield idx
                idx += 1

        if not update_properties:
            update_properties = self.update_properties

        super().__init__(
            load_size=load_size,
            update_tally=tally(),
            updates_per_reload=updates_per_reload,
            index=kwargs.pop("index", False)
            or (lambda load_size: np.random.randint(load_size)),
            number_of_updates=get_number_of_updates,
            update_properties=lambda: update_properties,
            **kwargs
        )

    def _process_and_get(self, *args, update_properties=None, **kwargs):

        # Loads a result from storage
        if self.feature and (
            not hasattr(self, "cache")
            or kwargs["update_tally"] - self.last_update >= kwargs["updates_per_reload"]
        ):
            if isinstance(self.feature, list):
                self.cache = [feature.resolve() for feature in self.feature]
            else:
                self.cache = self.feature.resolve()
            self.last_update = kwargs["update_tally"]

        if not self.feature:
            image_list_of_lists = args[0]
        else:
            image_list_of_lists = self.cache

        if not isinstance(image_list_of_lists, list):
            image_list_of_lists = [image_list_of_lists]

        new_list_of_lists = []
        # Calls get

        np.random.seed(kwargs["hash_key"][0])

        for image_list in image_list_of_lists:
            if isinstance(self.feature, list):
                # If multiple features, ensure consistent rng
                np.random.seed(kwargs["hash_key"][0])

            if isinstance(image_list, list):
                new_list_of_lists.append(
                    [
                        [
                            Image(
                                self.get(Image(image), **kwargs)
                            ).merge_properties_from(image)
                            for image in image_list
                        ]
                    ]
                )
            else:
                new_list_of_lists.append(
                    Image(self.get(Image(image_list), **kwargs)).merge_properties_from(
                        image_list
                    )
                )

        if update_properties:
            if not isinstance(new_list_of_lists, list):
                new_list_of_lists = [new_list_of_lists]
            for image_list in new_list_of_lists:
                if not isinstance(image_list, list):
                    image_list = [image_list]
                for image in image_list:
                    image.properties = [dict(prop) for prop in image.properties]
                    update_properties(image, **kwargs)

        return new_list_of_lists

    def _update(self, **kwargs):
        super()._update(**kwargs)
        if self.feature and not self.number_of_updates.current_value:
            if isinstance(self.feature, Feature):
                self.feature._update(**kwargs)
            elif isinstance(self.feature, list):
                [feature._update(**kwargs) for feature in self.feature]

    def update_properties(*args, **kwargs):
        pass


class PreLoad(Augmentation):
    """Simple storage with no augmentation.

    Parameters
    ----------
    feature : Feature
        The parent feature.
    load_size : int
        Number of results to resolve from the parent feature.
    updates_per_reload : int
        Number of times `.update()` is called before resolving new results
        from the parent feature.
    index : int
        The index of the image to grab from storage. By default a random image.
    update_properties : Callable or None
        Function called on the output of the method `.get()`. Overrides
        the default behaviour, allowing full control over how to update
        the properties of the output to account for the augmentation.

    """

    def get(self, image, **kwargs):
        return image


class FlipLR(Augmentation):
    """Flips images left-right.

    Updates all properties called "position" to flip the second index.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, load_size=1, updates_per_reload=2, **kwargs)

    def get(self, image, number_of_updates, **kwargs):
        if number_of_updates % 2:
            image = np.fliplr(image)
        return image

    def update_properties(self, image, number_of_updates, **kwargs):
        if number_of_updates % 2:
            for prop in image.properties:
                if "position" in prop:
                    position = prop["position"]
                    new_position = (
                        position[0],
                        image.shape[1] - position[1] - 1,
                        *position[2:],
                    )
                    prop["position"] = new_position


class FlipUD(Augmentation):
    """Flips images up-down.

    Updates all properties called "position" by flipping the first index.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, load_size=1, updates_per_reload=2, **kwargs)

    def get(self, image, number_of_updates=0, **kwargs):
        if number_of_updates % 2:
            image = np.flipud(image)
        return image

    def update_properties(self, image, number_of_updates, **kwargs):
        if number_of_updates % 2:
            for prop in image.properties:
                if "position" in prop:
                    position = prop["position"]
                    new_position = (image.shape[0] - position[0] - 1, *position[1:])
                    prop["position"] = new_position


class FlipDiagonal(Augmentation):
    """Flips images along the main diagonal.

    Updates all properties called "position" by swapping the first and second index.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, load_size=1, updates_per_reload=2, **kwargs)

    def get(self, image, number_of_updates, axes=(1, 0, 2), **kwargs):
        if number_of_updates % 2:
            image = np.transpose(image, axes=axes)
        return image

    def update_properties(self, image, number_of_updates, **kwargs):
        if number_of_updates % 2:
            for prop in image.properties:
                if "position" in prop:
                    position = prop["position"]
                    new_position = (position[1], position[0], *position[2:])
                    prop["position"] = new_position


from deeptrack.utils import get_kwarg_names
import warnings

import scipy.ndimage as ndimage
import deeptrack.utils as utils


class Affine(Augmentation):
    """
    Augmenter to apply affine transformations to images.

    Affine transformations involve:

        - Translation
        - Scaling
        - Rotation
        - Shearing

    Some transformations involve interpolations between several pixels
    of the input image to generate output pixel values. The parameter `order`
    deals with the method of interpolation used for this.


    Parameters
    ----------
    scale : number or tuple of number or list of number or dict {"x": number, "y": number}
        Scaling factor to use, where ``1.0`` denotes "no change" and
        ``0.5`` is zoomed out to ``50`` percent of the original size.
        If two values are provided (using tuple, list, or dict), the two first dimensions of the input are scaled individually.

    translate : number or tuple of number or list of number or dict {"x": number, "y": number}
        Translation in pixels.

    translate_px : number or tuple of number or list of number or dict {"x": number, "y": number}
        DEPRECATED, use translate.

    rotate : number
        Rotation in radians, i.e. Rotation happens around the *center* of the
        image.

    shear : number
        Shear in radians. Values in the range (-pi/4, pi/4) are common


    order : int
        Interpolation order to use. Same meaning as in ``skimage``:

            * ``0``: ``Nearest-neighbor``
            * ``1``: ``Bi-linear`` (default)
            * ``2``: ``Bi-quadratic`` (not recommended by skimage)
            * ``3``: ``Bi-cubic``
            * ``4``: ``Bi-quartic``
            * ``5``: ``Bi-quintic``
    """

    def __init__(
        self,
        scale=1,
        translate=None,
        translate_px=0,
        rotate=0,
        shear=0,
        order=1,
        cval=0,
        mode="reflect",
        **kwargs
    ):
        if translate is None:
            translate = translate_px
        super().__init__(
            scale=scale,
            translate=translate,
            translate_px=translate,
            rotate=rotate,
            shear=shear,
            order=order,
            cval=cval,
            mode=mode,
            **kwargs
        )

    def _process_properties(self, properties):

        properties = super()._process_properties(properties)

        # Make translate tuple
        translate = properties["translate"]
        if isinstance(translate, (float, int)):
            translate = (translate, translate)
        if isinstance(translate, dict):
            translate = (translate["x"], translate["y"])
        properties["translate"] = translate

        # Make scale tuple
        scale = properties["scale"]
        if isinstance(scale, (float, int)):
            scale = (scale, scale)
        if isinstance(scale, dict):
            scale = (scale["x"], scale["y"])
        properties["scale"] = scale

        return properties

    def get(self, image, scale, translate, rotate, shear, **kwargs):

        assert (
            image.ndim == 2 or image.ndim == 3
        ), "Affine only supports 2-dimensional or 3-dimension inputs."

        dx, dy = translate
        fx, fy = scale

        cr = np.cos(rotate)
        sr = np.sin(rotate)

        k = np.tan(shear)

        scale_map = np.array([[1 / fx, 0], [0, 1 / fy]])
        rotation_map = np.array([[cr, sr], [-sr, cr]])
        shear_map = np.array([[1, 0], [-k, 1]])

        mapping = scale_map @ rotation_map @ shear_map

        shape = image.shape
        center = np.array(shape[:2]) / 2

        d = center - np.dot(mapping, center) - np.array([dy, dx])

        # Clean up kwargs
        kwargs.pop("input", False)
        kwargs.pop("matrix", False)
        kwargs.pop("offset", False)
        kwargs.pop("output", False)

        # Call affine_transform
        if image.ndim == 2:
            utils.safe_call(
                ndimage.affine_transform,
                input=image,
                output=image,
                matrix=mapping,
                offset=d,
                **kwargs
            )

        elif image.ndim == 3:
            for z in range(shape[-1]):
                image[:, :, z] = utils.safe_call(
                    ndimage.affine_transform,
                    input=image[:, :, z],
                    matrix=mapping,
                    offset=d,
                    **kwargs
                )

        # Map positions
        inverse_mapping = np.linalg.inv(mapping)
        for prop in image.properties:
            if "position" in prop:
                position = np.array(prop["position"])
                prop["position"] = np.array(
                    (
                        *(
                            (
                                inverse_mapping
                                @ (position[:2] - center + np.array([dy, dx]))
                                + center
                            )
                        ),
                        *position[3:],
                    )
                )

        return image


class ElasticTransformation(Augmentation):
    """Transform images by moving pixels locally around using displacement fields.

    The augmenter creates a random distortion field using `alpha` and `sigma`, which define
    the strength and smoothness of the field respectively. These are used to transform the
    input locally.

    .. Note:
        This augmentation does not currently update the position property of the image,
        meaning that it is not recommended to use it if the network label is
        derived from the position properties of the resulting image.

    For a detailed explanation, see ::

        Simard, Steinkraus and Platt
        Best Practices for Convolutional Neural Networks applied to Visual
        Document Analysis
        in Proc. of the International Conference on Document Analysis and
        Recognition, 2003


    Parameters
    ----------
    alpha : number
        Strength of the distortion field. Common values are in the range (10, 100)

    sigma : number
        Standard deviation of the gaussian kernel used to smooth the distortion
        fields. Common values are in the range (1, 10)

    ignore_last_dim : bool
        Whether to skip creating a distortion field for the last dimension.
        This is often desired if the last dimension is a channel dimension (such as
        a color image.) In that case, the three channels are transformed identically
        and do not `bleed` into eachother.


    order : int
        Interpolation order to use. Takes integers from 0 to 5

            * ``0``: ``Nearest-neighbor``
            * ``1``: ``Bi-linear`` (default)
            * ``2``: ``Bi-quadratic`` (not recommended by skimage)
            * ``3``: ``Bi-cubic``
            * ``4``: ``Bi-quartic``
            * ``5``: ``Bi-quintic``


    cval : number
        The constant intensity value used to fill in new pixels.
        This value is only used if `mode` is set to ``constant``.

    mode : str
        Parameter that defines newly created pixels.
        May take the same values as in :func:`scipy.ndimage.map_coordinates`,
        i.e. ``constant``, ``nearest``, ``reflect`` or ``wrap``.

    """

    def __init__(
        self,
        alpha=20,
        sigma=2,
        ignore_last_dim=True,
        order=3,
        cval=0,
        mode="constant",
        **kwargs
    ):
        super().__init__(
            alpha=alpha,
            sigma=sigma,
            ignore_last_dim=ignore_last_dim,
            order=order,
            cval=cval,
            mode=mode,
            **kwargs
        )

    def get(self, image, sigma, alpha, ignore_last_dim, **kwargs):

        shape = image.shape

        if ignore_last_dim:
            shape = shape[:-1]

        deltas = []
        ranges = []
        coordinates = []

        for dim in shape:
            deltas.append(
                gaussian_filter(
                    (np.random.rand(*shape) * 2 - 1), sigma, mode="constant", cval=0
                )
                * alpha
            )

            ranges.append(np.arange(dim))

        grids = list(np.meshgrid(*ranges))

        for grid, delta in zip(grids, deltas):
            dDim = grid + delta
            coordinates.append(np.reshape(dDim, (-1, 1)))

        if ignore_last_dim:
            for z in range(image.shape[-1]):
                image[..., z] = utils.safe_call(
                    map_coordinates,
                    input=image[..., z],
                    coordinates=coordinates,
                    **kwargs
                ).reshape(shape)
        else:
            image = utils.safe_call(
                map_coordinates, input=image, coordinates=coordinates, **kwargs
            ).reshape(shape)

        # TODO: implement interpolated coordinate mapping for property positions
        # for prop in image:
        #     if "position" in prop:

        return image


class Crop(Augmentation):
    """Crops a regions of an image.

    Parameters
    ----------
    feature : feature or list of features
        Feature(s) to augment.
    crop : int or tuple of int or list of int or Callable[Image]->tuple of ints
        Number of pixels to remove or retain (depending in `crop_mode`)
        If a tuple or list, it is assumed to be per axis.
        Can also be a function that returns any of the other types.
    crop_mode : str {"remove", "retain"}
        How the `crop` argument is interpreted. If "remove", then
        crop
    corner : tuple of ints or Callable[Image]->tuple of ints or "random"
        Top left corner of the cropped region. Can be a tuple of ints,
        a function that returns a tuple of ints or the string random.
        If corner is placed so that the cropping cannot be performed,
        the modulo of the corner with the allowed region is used.

    """

    def __init__(
        self, *args, crop=(64, 64), crop_mode="retain", corner="random", **kwargs
    ):
        super().__init__(*args, crop=crop, crop_mode=crop_mode, corner=corner, **kwargs)

    def get(self, image, corner, crop, crop_mode, **kwargs):

        # Get crop argument
        if callable(crop):
            crop = crop(image)
        if isinstance(crop, int):
            crop = (crop,) * image.ndim

        # Get amount to crop from image
        if crop_mode == "retain":
            crop_amount = np.array(image.shape) - np.array(crop)
        elif crop_mode == "remove":
            crop_amount = np.array(crop)
        else:
            raise ValueError("Unrecognized crop_mode {0}".format(crop_mode))

        # Contain within image
        crop_amount = np.amax((np.array(crop_amount), [0] * image.ndim), axis=0)
        crop_amount = np.amin((np.array(image.shape) - 1, crop_amount), axis=0)
        # Get corner of crop
        if corner == "random":
            # Ensure seed is consistent
            slice_start = np.random.randint(
                [0] * crop_amount.size,
                crop_amount + 1,
            )
        elif callable(corner):
            slice_start = corner(image)
        else:
            slice_start = corner

        # Ensure compatible with image
        slice_start = [c % (m + 1) for c, m in zip(slice_start, crop_amount)]
        slice_end = [
            a - c + s for a, s, c in zip(image.shape, slice_start, crop_amount)
        ]

        slices = tuple(
            [
                slice(slice_start_i, slice_end_i)
                for slice_start_i, slice_end_i in zip(slice_start, slice_end)
            ]
        )
        cropped_image = image[slices]

        # Update positions
        cropped_image.properties = [dict(prop) for prop in image.properties]
        for prop in cropped_image.properties:
            if "position" in prop:
                position = np.array(prop["position"])
                try:
                    position[0:2] -= np.array(slice_start)[0:2]
                    prop["position"] = position
                except IndexError:
                    pass

        return cropped_image


class CropToMultiplesOf(Crop):
    """Crop images down until their height/width is a multiple of a value.

    Parameters
    ----------
    multiple : int or tuple of (int or None)
        Images will be cropped down until their width is a multiple of
        this value. If a tuple, it is assumed to be a multiple per axis.
        A value of None or -1 indicates to skip that axis.

    """

    def __init__(self, multiple=1, corner="random", **kwargs):
        kwargs.pop("crop", False)
        kwargs.pop("crop_mode", False)

        def image_to_crop(image):
            shape = image.shape
            multiple = self.multiple.current_value

            if not isinstance(multiple, (list, tuple, np.ndarray)):
                multiple = (multiple,) * image.ndim
            new_shape = list(shape)
            idx = 0
            for dim, mul in zip(shape, multiple):
                if mul is not None and mul is not -1:
                    new_shape[idx] = int((dim // mul) * mul)
                idx += 1
            return new_shape

        super().__init__(
            multiple=multiple,
            corner=corner,
            crop=lambda: image_to_crop,
            crop_mode="retain",
            **kwargs
        )


# TODO: add padding
# TODO: add resizing by rescaling


""" Features that aberrate and modify pupil functions.

Classes
-------
Aberration
    Base abstract class
GaussianApodization
    Introduces pupil apodization.
Zernike
    Introduces a Zernike phase aberration.
Piston
    Zernike polynomial with n=0, m=0.
VerticalTilt
    Zernike polynomial with n=1, m=-1.
HorizontalTilt
    Zernike polynomial with n=1, m=1.
ObliqueAstigmatism
    Zernike polynomial with n=2, m=-2.
Defocus
    Zernike polynomial with n=2, m=0.
Astigmatism
    Zernike polynomial with n=2, m=2.
ObliqueTrefoil
    Zernike polynomial with n=3, m=-3.
VerticalComa
    Zernike polynomial with n=3, m=-1.
HorizontalComa
    Zernike polynomial with n=3, m=1.
Trefoil
    Zernike polynomial with n=3, m=3.
SphericalAberration
    Zernike polynomial with n=4, m=0.
"""

import numpy as np
from deeptrack.features import Feature
from deeptrack.utils import as_list


class Aberration(Feature):
    """Base abstract class.

    Ensures that the method `.get()` receives rho and theta as optional
    arguments, describing the polar coordinates of each pixel in the image
    scaled so that rho is 1 at the edge of the pupil.
    """

    __distributed__ = True

    # Adds rho and theta of the pupil to the input.
    def _process_and_get(self, image_list, **kwargs):
        new_list = []
        for image in image_list:
            x = np.arange(image.shape[0]) - image.shape[0] / 2
            y = np.arange(image.shape[1]) - image.shape[1] / 2
            X, Y = np.meshgrid(x, y)
            rho = np.sqrt(X ** 2 + Y ** 2)
            rho /= np.max(rho[image != 0])
            theta = np.arctan2(Y, X)

            new_list += super()._process_and_get(
                [image], rho=rho, theta=theta, **kwargs
            )
        return new_list


# AMPLITUDE ABERRATIONS


class GaussianApodization(Aberration):
    """Introduces pupil apodization.

    Decreases the amplitude of the pupil at high frequencies according
    to a Gaussian distribution.

    Parameters
    ----------
    sigma : float
        The standard deviation of the apodization. The edge of the pupil
        is at one deviation from the center.
    offset : (float, float)
        Offsets the center of the gaussian.

    """

    def __init__(self, sigma=1, offset=(0, 0), **kwargs):
        super().__init__(sigma=sigma, offset=offset, **kwargs)

    def get(self, pupil, offset, sigma, rho, **kwargs):
        if offset != (0, 0):
            x = np.arange(pupil.shape[0]) - pupil.shape[0] / 2 - offset[0]
            y = np.arange(pupil.shape[1]) - pupil.shape[1] / 2 - offset[1]
            X, Y = np.meshgrid(x, y)
            rho = np.sqrt(X ** 2 + Y ** 2)
            rho /= np.max(rho[pupil != 0])
            rho[rho > 1] = np.inf

        pupil = pupil * np.exp(-((rho / sigma) ** 2))
        return pupil


# PHASE ABERRATIONS


class Zernike(Aberration):
    """Introduces a Zernike phase aberration.

    Calculates the Zernike polynomial defined by the numbers `n` and `m` at
    each pixel in the pupil, multiplies it by `coefficient`, and adds the
    result to the phase of the pupil.

    If `n`, `m` and `coefficient` are lists of equal lengths, sum the
    Zernike polynomials corresponding to each set of values in these lists
    before adding them to the phase.

    Parameters
    ----------
    n, m : int or list of ints
        The zernike polynomial numbers.
    coefficient : float or list of floats
        The coefficient of the polynomial
    """

    def __init__(self, n, m, coefficient=1, **kwargs):
        super().__init__(n=n, m=m, coefficient=coefficient, **kwargs)

    def get(self, pupil, rho, theta, n, m, coefficient, **kwargs):
        m_list = as_list(m)
        n_list = as_list(n)
        coefficients = as_list(coefficient)

        assert len(m_list) == len(n_list), "The number of indices need to match"
        assert len(m_list) == len(
            coefficients
        ), "The number of indices need to match the number of coefficients"

        pupil_bool = pupil != 0

        rho = rho[pupil_bool]
        theta = theta[pupil_bool]

        Z = 0

        for n, m, coefficient in zip(n_list, m_list, coefficients):
            if (n - m) % 2 or coefficient == 0:
                continue

            R = 0
            for k in range((n - np.abs(m)) // 2 + 1):
                R += (
                    (-1) ** k
                    * np.math.factorial(n - k)
                    / (
                        np.math.factorial(k)
                        * np.math.factorial((n - m) // 2 - k)
                        * np.math.factorial((n + m) // 2 - k)
                    )
                    * rho ** (n - 2 * k)
                )

            if m > 0:
                R = R * np.cos(m * theta) * (np.sqrt(2 * n + 2) * coefficient)
            elif m < 0:
                R = R * np.sin(-m * theta) * (np.sqrt(2 * n + 2) * coefficient)
            else:
                R = R * (np.sqrt(n + 1) * coefficient)

            Z += R

        phase = np.exp(1j * Z)

        pupil[pupil_bool] *= phase

        return pupil


# COMMON ABERRATIONS


class Piston(Zernike):
    """Zernike polynomial with n=0, m=0.

    Parameters
    ----------
    coefficient : float
        The coefficient of the polynomial
    """

    def __init__(self, *args, coefficient=1, **kwargs):
        super().__init__(*args, n=0, m=0, coefficient=coefficient, **kwargs)


class VerticalTilt(Zernike):
    """Zernike polynomial with n=1, m=-1.

    Parameters
    ----------
    coefficient : float
        The coefficient of the polynomial
    """

    def __init__(self, *args, coefficient=1, **kwargs):
        super().__init__(*args, n=1, m=-1, coefficient=coefficient, **kwargs)


class HorizontalTilt(Zernike):
    """Zernike polynomial with n=1, m=1.

    Parameters
    ----------
    coefficient : float
        The coefficient of the polynomial
    """

    def __init__(self, *args, coefficient=1, **kwargs):
        super().__init__(*args, n=1, m=1, coefficient=coefficient, **kwargs)


class ObliqueAstigmatism(Zernike):
    """Zernike polynomial with n=2, m=-2.

    Parameters
    ----------
    coefficient : float
        The coefficient of the polynomial
    """

    def __init__(self, *args, coefficient=1, **kwargs):
        super().__init__(*args, n=2, m=-2, coefficient=coefficient, **kwargs)


class Defocus(Zernike):
    """Zernike polynomial with n=2, m=0.

    Parameters
    ----------
    coefficient : float
        The coefficient of the polynomial
    """

    def __init__(self, *args, coefficient=1, **kwargs):
        super().__init__(*args, n=2, m=0, coefficient=coefficient, **kwargs)


class Astigmatism(Zernike):
    """Zernike polynomial with n=2, m=2.

    Parameters
    ----------
    coefficient : float
        The coefficient of the polynomial
    """

    def __init__(self, *args, coefficient=1, **kwargs):
        super().__init__(*args, n=2, m=2, coefficient=coefficient, **kwargs)


class ObliqueTrefoil(Zernike):
    """Zernike polynomial with n=3, m=-3.

    Parameters
    ----------
    coefficient : float
        The coefficient of the polynomial
    """

    def __init__(self, *args, coefficient=1, **kwargs):
        super().__init__(*args, n=3, m=-3, coefficient=coefficient, **kwargs)


class VerticalComa(Zernike):
    """Zernike polynomial with n=3, m=-1.

    Parameters
    ----------
    coefficient : float
        The coefficient of the polynomial
    """

    def __init__(self, *args, coefficient=1, **kwargs):
        super().__init__(*args, n=3, m=-1, coefficient=coefficient, **kwargs)


class HorizontalComa(Zernike):
    """Zernike polynomial with n=3, m=1.

    Parameters
    ----------
    coefficient : float
        The coefficient of the polynomial
    """

    def __init__(self, *args, coefficient=1, **kwargs):
        super().__init__(*args, n=3, m=1, coefficient=coefficient, **kwargs)


class Trefoil(Zernike):
    """Zernike polynomial with n=3, m=3.

    Parameters
    ----------
    coefficient : float
        The coefficient of the polynomial
    """

    def __init__(self, *args, coefficient=1, **kwargs):
        super().__init__(*args, n=3, m=3, coefficient=coefficient, **kwargs)


class SphericalAberration(Zernike):
    """Zernike polynomial with n=4, m=0.

    Parameters
    ----------
    coefficient : float
        The coefficient of the polynomial
    """

    def __init__(self, *args, coefficient=1, **kwargs):
        super().__init__(*args, n=4, m=0, coefficient=coefficient, **kwargs)



import sys

sys.path.append("..")  # Adds the module to path

import unittest

import deeptrack.features as features

import numpy as np
from deeptrack.image import Image


class TestFeatures(unittest.TestCase):
    def test_Feature_1(self):
        class FeatureConcreteClass(features.Feature):
            __distributed__ = False

            def get(self, *args, **kwargs):
                image = np.ones((2, 3))
                return image

        feature = FeatureConcreteClass()
        feature.update()
        output_image = feature.resolve()
        self.assertIsInstance(output_image, Image)
        self.assertEqual(output_image.size, 6)

    def test_Feature_2(self):
        class FeatureAddValue(features.Feature):
            def get(self, image, value_to_add=0, **kwargs):
                image = image + value_to_add
                return image

        feature = FeatureAddValue(value_to_add=1)
        feature.update()
        input_image = np.zeros((1, 1))
        output_image = feature.resolve(input_image)
        self.assertEqual(output_image, 1)
        self.assertListEqual(
            output_image.get_property("value_to_add", get_one=False), [1]
        )
        output_image = feature.resolve(output_image)
        self.assertEqual(output_image, 2)
        self.assertListEqual(
            output_image.get_property("value_to_add", get_one=False), [1, 1]
        )

    def test_Feature_with_dummy_property(self):
        class FeatureConcreteClass(features.Feature):
            __distributed__ = False

            def get(self, *args, **kwargs):
                image = np.ones((2, 3))
                return image

        feature = FeatureConcreteClass(dummy_property="foo")
        feature.update()
        output_image = feature.resolve()
        self.assertListEqual(
            output_image.get_property("dummy_property", get_one=False), ["foo"]
        )

    def test_Feature_plus_1(self):
        class FeatureAddValue(features.Feature):
            def get(self, image, value_to_add=0, **kwargs):
                image = image + value_to_add
                return image

        feature1 = FeatureAddValue(value_to_add=1)
        feature2 = FeatureAddValue(value_to_add=2)
        feature = feature1 + feature2
        feature.update()
        input_image = np.zeros((1, 1))
        output_image = feature.resolve(input_image)
        self.assertEqual(output_image, 3)
        self.assertListEqual(
            output_image.get_property("value_to_add", get_one=False), [1, 2]
        )
        self.assertEqual(output_image.get_property("value_to_add", get_one=True), 1)

    def test_Feature_plus_2(self):
        class FeatureAddValue(features.Feature):
            def get(self, image, value_to_add=0, **kwargs):
                image = image + value_to_add
                return image

        class FeatureMultiplyByValue(features.Feature):
            def get(self, image, value_to_multiply=0, **kwargs):
                image = image * value_to_multiply
                return image

        feature1 = FeatureAddValue(value_to_add=1)
        feature2 = FeatureMultiplyByValue(value_to_multiply=10)
        input_image = np.zeros((1, 1))
        feature12 = feature1 + feature2
        feature12.update()
        output_image12 = feature12.resolve(input_image)
        self.assertEqual(output_image12, 10)
        feature21 = feature2 + feature1
        output_image21 = feature21.resolve(input_image)
        self.assertEqual(output_image21, 1)

    def test_Feature_plus_3(self):
        class FeatureAppendImageOfShape(features.Feature):
            __distributed__ = False
            __list_merge_strategy__ = features.MERGE_STRATEGY_APPEND

            def get(self, *args, shape, **kwargs):
                image = np.zeros(shape)
                return image

        feature1 = FeatureAppendImageOfShape(shape=(1, 1))
        feature2 = FeatureAppendImageOfShape(shape=(2, 2))
        feature12 = feature1 + feature2
        feature12.update()
        output_image = feature12.resolve()
        self.assertIsInstance(output_image, list)
        self.assertIsInstance(output_image[0], Image)
        self.assertIsInstance(output_image[1], Image)
        self.assertEqual(output_image[0].shape, (1, 1))
        self.assertEqual(output_image[1].shape, (2, 2))

    def test_Feature_times_1(self):
        class FeatureAddValue(features.Feature):
            def get(self, image, value_to_add=0, **kwargs):
                image = image + value_to_add
                return image

        input_image = np.zeros((1, 1))
        feature0 = FeatureAddValue(value_to_add=1) * 0
        feature0.update()
        output_image0 = feature0.resolve(input_image)
        self.assertEqual(output_image0, 0)
        feature1 = FeatureAddValue(value_to_add=1) * 1
        feature1.update()
        output_image1 = feature1.resolve(input_image)
        self.assertEqual(output_image1, 1)
        feature05 = FeatureAddValue(value_to_add=1) * 0.5
        for _ in range(100):
            feature05.update()
            output_image05 = feature05.resolve(input_image)
            self.assertTrue(output_image05[0, 0] == 0 or output_image05[0, 0] == 1)

    def test_Feature_exp_1(self):
        class FeatureAddValue(features.Feature):
            def get(self, image, value_to_add=0, **kwargs):
                image = image + value_to_add
                return image

        feature = FeatureAddValue(value_to_add=1) ** 10
        feature.update()
        input_image = np.zeros((1, 1))
        output_image = feature.resolve(input_image)
        self.assertEqual(output_image, 10)

    def test_Feature_property_memorability(self):
        class FeatureWithForgettableProperties(features.Feature):
            __property_memorability__ = 2

            def get(self, image, forgettable_property, **kwargs):
                return image

        feature = FeatureWithForgettableProperties(forgettable_property=1)
        feature.update()
        input_image = np.zeros((1, 1))
        output_image = feature.resolve(input_image)
        self.assertIsNone(
            output_image.get_property("forgettable_property", default=None)
        )
        output_image = feature.resolve(input_image, property_memorability=2)
        self.assertEqual(
            output_image.get_property("forgettable_property", default=None), 1
        )

    def test_Feature_with_overruled_property(self):
        class FeatureAddValue(features.Feature):
            def get(self, image, value_to_add=0, **kwargs):
                image = image + value_to_add
                return image

        feature = FeatureAddValue(value_to_add=1)
        feature.update()
        input_image = np.zeros((1, 1))
        output_image = feature.resolve(input_image, value_to_add=10)
        self.assertEqual(output_image, 10)
        self.assertListEqual(
            output_image.get_property("value_to_add", get_one=False), [10]
        )
        self.assertEqual(output_image.get_property("value_to_add", get_one=True), 10)

    def test_nested_Duplicate(self):
        for _ in range(5):
            A = features.DummyFeature(
                a=lambda: np.random.randint(100) * 1000,
            )
            B = features.DummyFeature(
                a=A.a,
                b=lambda a: a + np.random.randint(10) * 100,
            )
            C = features.DummyFeature(
                b=B.b,
                c=lambda b: b + np.random.randint(10) * 10,
            )
            D = features.DummyFeature(
                c=C.c,
                d=lambda c: c + np.random.randint(10) * 1,
            )

            for _ in range(5):
                AB = (A + (B + (C + D) ** 2) ** 2) ** 6
                output = AB.update().resolve(0)
                al = output.get_property("a", get_one=False)[::3]
                bl = output.get_property("b", get_one=False)[::3]
                cl = output.get_property("c", get_one=False)[::2]
                dl = output.get_property("d", get_one=False)[::1]

                self.assertFalse(all(a == al[0] for a in al))
                self.assertFalse(all(b == bl[0] for b in bl))
                self.assertFalse(all(c == cl[0] for c in cl))
                self.assertFalse(all(d == dl[0] for d in dl))
                for ai, a in enumerate(al):
                    for bi, b in list(enumerate(bl))[ai * 2 : (ai + 1) * 2]:
                        self.assertIn(b - a, range(0, 1000))
                        for ci, c in list(enumerate(cl))[bi * 2 : (bi + 1) * 2]:
                            self.assertIn(c - b, range(0, 100))
                            self.assertIn(dl[ci] - c, range(0, 10))

    def test_LambdaDependence(self):
        A = features.DummyFeature(a=1, b=2, c=3)

        B = features.DummyFeature(
            key="a",
            prop=lambda key: A.a if key == "a" else (A.b if key == "b" else A.c),
        )

        B.update()
        self.assertEqual(B.prop.current_value, 1)
        B.update(key="a")
        self.assertEqual(B.prop.current_value, 1)
        B.update(key="b")
        self.assertEqual(B.prop.current_value, 2)
        B.update(key="c")
        self.assertEqual(B.prop.current_value, 3)

    def test_LambdaDependenceTwice(self):
        A = features.DummyFeature(a=1, b=2, c=3)

        B = features.DummyFeature(
            key="a",
            prop=lambda key: A.a if key == "a" else (A.b if key == "b" else A.c),
            prop2=lambda prop: prop * 2,
        )

        B.update()
        self.assertEqual(B.prop2.current_value, 2)
        B.update(key="a")
        self.assertEqual(B.prop2.current_value, 2)
        B.update(key="b")
        self.assertEqual(B.prop2.current_value, 4)
        B.update(key="c")
        self.assertEqual(B.prop2.current_value, 6)

    def test_LambdaDependenceOtherFeature(self):
        A = features.DummyFeature(a=1, b=2, c=3)

        B = features.DummyFeature(
            key="a",
            prop=lambda key: A.a if key == "a" else (A.b if key == "b" else A.c),
            prop2=lambda prop: prop * 2,
        )

        C = features.DummyFeature(B_prop=B.prop2, prop=lambda B_prop: B_prop * 2)

        C.update()
        self.assertEqual(C.prop.current_value, 4)
        C.update(key="a")
        self.assertEqual(C.prop.current_value, 4)
        C.update(key="b")
        self.assertEqual(C.prop.current_value, 8)
        C.update(key="c")
        self.assertEqual(C.prop.current_value, 12)


if __name__ == "__main__":
    unittest.main()